import pytest
from mento.rebar import Rebar
from mento.beam import RectangularBeam
from mento.material import Concrete_ACI_318_19, SteelBar
from mento.units import psi, kip, mm, inch, ksi, cm, MPa, kN
from mento.forces import Forces
from mento.node import Node


@pytest.fixture()
def beam_example_imperial() -> RectangularBeam:
    concrete = Concrete_ACI_318_19(name="C4", f_c=4000 * psi)
    steelBar = SteelBar(name="ADN 420", f_y=60 * ksi)
    custom_settings = {
        "clear_cover": 1.5 * inch,
        "stirrup_diameter_ini": 0.5 * inch,
        "longitudinal_diameter_ini": 1 * inch,
    }
    section = RectangularBeam(
        label="V-10x16",
        concrete=concrete,
        steel_bar=steelBar,
        width=10 * inch,
        height=16 * inch,
        settings=custom_settings,
    )
    return section


@pytest.fixture()
def beam_example_metric() -> RectangularBeam:
    concrete = Concrete_ACI_318_19(name="H30", f_c=30 * MPa)
    steelBar = SteelBar(name="ADN 420", f_y=420 * MPa)
    custom_settings = {"clear_cover": 30 * mm, "stirrup_diameter_ini": 8 * mm}
    section = RectangularBeam(
        label="V 20x50",
        concrete=concrete,
        steel_bar=steelBar,
        width=20 * cm,
        height=50 * cm,
        settings=custom_settings,
    )
    return section


# Tests
def test_beam_longitudinal_rebar_ACI_318_19_metric(
    beam_example_metric: RectangularBeam,
) -> None:
    A_s_req = 5 * cm**2
    beam_rebar = Rebar(beam_example_metric)
    beam_rebar.longitudinal_rebar_ACI_318_19(A_s_req=A_s_req)
    best_design = beam_rebar.longitudinal_rebar_design

    assert best_design is not None
    assert best_design["n_1"] == 2
    assert best_design["d_b1"] == 16 * mm
    assert best_design["n_2"] == 1
    assert best_design["d_b2"] == 12 * mm
    assert best_design["n_3"] == 0
    assert best_design["d_b3"] is None
    assert best_design["n_4"] == 0
    assert best_design["d_b3"] is None
    assert best_design["total_as"].magnitude == pytest.approx(5.15, rel=1e-3)
    assert best_design["total_bars"] == 3
    assert best_design["clear_spacing"].magnitude == pytest.approx(40, rel=1e-3)


# def test_beam_longitudinal_rebar_ACI_318_19_imperial(beam_example_imperial: RectangularBeam) -> None:
#     # This test will ensure that the longitudinal rebar design works correctly for beams defined in imperial units.
#     A_s_req = 1.5 * inch**2
#     beam_rebar = Rebar(beam_example_imperial)
#     beam_rebar.longitudinal_rebar_ACI_318_19(A_s_req=A_s_req)
#     best_design = beam_rebar.longitudinal_rebar_design

#     assert best_design is not None
#     assert best_design['n_1'] == 2
#     assert best_design['d_b1'] == 0.625 * inch
#     assert best_design['n_2'] == 1
#     assert best_design['d_b2'] == 0.5 * inch
#     assert best_design['n_3'] == 0
#     assert best_design['d_b3'] is None
#     assert best_design['n_4'] == 0
#     assert best_design['d_b3'] is None
#     assert best_design['total_as'].magnitude == pytest.approx(1.53, rel=1e-3)
#     assert best_design['total_bars'] == 3
#     assert best_design['clear_spacing'].magnitude == pytest.approx(1.5, rel=1e-3)


def test_beam_longitudinal_rebar_min_diameter(
    beam_example_metric: RectangularBeam,
) -> None:
    # This test ensures that the minimum rebar diameter constraint is respected.
    A_s_req = 2 * cm**2
    beam_rebar = Rebar(beam_example_metric)
    beam_rebar.longitudinal_rebar_ACI_318_19(A_s_req=A_s_req)
    best_design = beam_rebar.longitudinal_rebar_design

    assert best_design is not None
    assert best_design["d_b1"] >= 10 * mm
    assert best_design["d_b2"] >= 10 * mm if best_design["d_b2"] is not None else True


def test_beam_longitudinal_rebar_max_bars_per_layer(
    beam_example_metric: RectangularBeam,
) -> None:
    # This test ensures that the maximum number of bars per layer constraint is respected.
    A_s_req = 10 * cm**2
    beam_rebar = Rebar(beam_example_metric)
    beam_rebar.longitudinal_rebar_ACI_318_19(A_s_req=A_s_req)
    best_design = beam_rebar.longitudinal_rebar_design

    assert best_design is not None
    assert best_design["n_1"] + best_design["n_2"] <= beam_rebar.max_bars_per_layer
    assert best_design["n_3"] + best_design["n_4"] <= beam_rebar.max_bars_per_layer


def test_beam_longitudinal_rebar_CIRSOC_201_25(
    beam_example_metric: RectangularBeam,
) -> None:
    # This test ensures that the longitudinal rebar design works correctly for different design codes.
    beam_example_metric.concrete.design_code = "CIRSOC 201-25"
    A_s_req = 5 * cm**2
    beam_rebar = Rebar(beam_example_metric)
    beam_rebar.longitudinal_rebar(A_s_req=A_s_req)
    best_design = beam_rebar.longitudinal_rebar_design

    assert best_design is not None
    assert best_design["n_1"] == 2
    assert best_design["d_b1"] == 16 * mm
    assert best_design["n_2"] == 1
    assert best_design["d_b2"] == 12 * mm
    assert best_design["n_3"] == 0
    assert best_design["d_b3"] is None
    assert best_design["n_4"] == 0
    assert best_design["d_b3"] is None
    assert best_design["total_as"].magnitude == pytest.approx(5.15, rel=1e-3)
    assert best_design["total_bars"] == 3
    assert best_design["clear_spacing"].magnitude == pytest.approx(40, rel=1e-3)


def test_beam_longitudinal_rebar_small_area(
    beam_example_metric: RectangularBeam,
) -> None:
    # This test ensures that the rebar design handles very small required areas correctly.
    A_s_req = 0.1 * cm**2
    beam_rebar = Rebar(beam_example_metric)
    beam_rebar.longitudinal_rebar_ACI_318_19(A_s_req=A_s_req)
    best_design = beam_rebar.longitudinal_rebar_design

    assert best_design is not None
    assert best_design["total_as"] >= A_s_req


def test_beam_longitudinal_rebar_large_area(
    beam_example_metric: RectangularBeam,
) -> None:
    # This test ensures that the rebar design handles very large required areas correctly.
    A_s_req = 50 * cm**2
    beam_rebar = Rebar(beam_example_metric)
    beam_rebar.longitudinal_rebar_ACI_318_19(A_s_req=A_s_req)
    best_design = beam_rebar.longitudinal_rebar_design

    assert best_design is not None
    assert best_design["n_1"] >= 2
    assert best_design["d_b1"] >= 8 * mm


def test_beam_longitudinal_rebar_zero_area_metric_ACI(
    beam_example_metric: RectangularBeam,
) -> None:
    A_s_req = 0 * cm**2
    beam_rebar = Rebar(beam_example_metric)
    beam_rebar.longitudinal_rebar_ACI_318_19(A_s_req=A_s_req)
    best_design = beam_rebar.longitudinal_rebar_design

    assert best_design is not None
    assert best_design["total_as"].magnitude == pytest.approx(
        1.5707, rel=1e-3
    )  # 2Ø10 minimum


# def test_beam_longitudinal_rebar_zero_area_imperial(beam_example_imperial: RectangularBeam) -> None:
#     A_s_req = 0 * inch**2
#     beam_rebar = Rebar(beam_example_imperial)
#     beam_rebar.longitudinal_rebar_ACI_318_19(A_s_req=A_s_req)
#     best_design = beam_rebar.longitudinal_rebar_design

#     assert best_design is not None
#     assert best_design['total_as'].magnitude == pytest.approx(1.5707,rel=1e-3) # 2#3 minimum


def test_beam_transverse_rebar_ACI_318_19_imperial(
    beam_example_imperial: RectangularBeam,
) -> None:
    f = Forces(V_z=37.727 * kip)
    beam_example_imperial.set_longitudinal_rebar_bot(n1=2, d_b1=0.625 * inch)
    N1 = Node(beam_example_imperial, forces=f)
    results = N1.design_shear()
    assert results is not None
    assert beam_example_imperial._stirrup_n == 1
    assert beam_example_imperial._stirrup_d_b == 3 / 8 * inch
    assert beam_example_imperial._stirrup_s_l == 5 * inch
    assert beam_example_imperial._A_v.to("cm**2/m").magnitude == pytest.approx(
        11.2214, rel=1e-3
    )


def test_beam_transverse_rebar_ACI_318_19_metric(
    beam_example_metric: RectangularBeam,
) -> None:
    # This test ensures that the transverse rebar design works correctly for beams defined in metric units.
    f = Forces(V_z=100 * kN)
    beam_example_metric.set_longitudinal_rebar_bot(n1=2, d_b1=16 * mm)
    N1 = Node(beam_example_metric, forces=f)
    results = N1.design_shear()

    assert results is not None
    assert beam_example_metric._stirrup_n == 1
    assert beam_example_metric._stirrup_d_b == 10 * mm
    assert beam_example_metric._stirrup_s_l == 22 * cm
    assert beam_example_metric._A_v.to("cm**2/m").magnitude == pytest.approx(
        7.14, rel=1e-3
    )


def test_beam_transverse_rebar_max_spacing_imperial(
    beam_example_imperial: RectangularBeam,
) -> None:
    # This test ensures that the maximum spacing constraints are respected.
    f = Forces(V_z=50 * kip)
    beam_example_imperial.set_longitudinal_rebar_bot(n1=2, d_b1=0.625 * inch)
    N1 = Node(beam_example_imperial, forces=f)
    results = N1.design_shear()

    assert results is not None
    assert beam_example_imperial._stirrup_s_l <= 24 * inch
    assert beam_example_imperial._stirrup_s_w <= 24 * inch


def test_beam_transverse_rebar_min_diameter(
    beam_example_metric: RectangularBeam,
) -> None:
    # This test ensures that the minimum rebar diameter constraint is respected for transverse reinforcement.
    f = Forces(V_z=50 * kN)
    beam_example_metric.set_longitudinal_rebar_bot(n1=2, d_b1=16 * mm)
    N1 = Node(beam_example_metric, forces=f)
    results = N1.design_shear()

    assert results is not None
    assert beam_example_metric._stirrup_d_b >= 8 * mm


def test_beam_transverse_rebar_CIRSOC_201_25(
    beam_example_metric: RectangularBeam,
) -> None:
    # This test ensures that the transverse rebar design works correctly for different design codes.
    beam_example_metric.concrete.design_code = "CIRSOC 201-25"
    f = Forces(V_z=100 * kN)
    beam_example_metric.set_longitudinal_rebar_bot(n1=2, d_b1=16 * mm)
    N1 = Node(beam_example_metric, forces=f)
    results = N1.design_shear()

    assert results is not None
    assert beam_example_metric._stirrup_n == 1
    assert beam_example_metric._stirrup_d_b == 6 * mm
    assert beam_example_metric._stirrup_s_l == 22 * cm
    assert beam_example_metric._A_v.to("cm**2/m").magnitude == pytest.approx(
        2.57, rel=1e-3
    )


# This is where pytest will collect the tests and run them
if __name__ == "__main__":
    pytest.main()



from typing import Union, Optional, List
import pandas as pd

from mento.section import Section
from mento.forces import Forces


class Node:
    """
    Node Class
    The `Node` class represents a structural node that is associated with a section
    (e.g., a beam) and can have forces applied to it. It provides methods for managing
    forces, performing structural checks, and designing for flexure and shear.
    Attributes:
        section (Section): The section (e.g., beam) associated with the node.
        forces (List[Forces]): A list of forces applied to the node.
    Methods:
        add_forces(forces: Union[Forces, List[Forces]]) -> None:
            Adds one or more forces to the node.
        get_forces_list() -> List[Forces]:
            Returns the list of forces applied to this node.
        clear_forces() -> None:
            Removes all forces from the node.
        check_flexure() -> pd.DataFrame:
            Performs a flexural check on the section using the applied forces.
        design_flexure() -> pd.DataFrame:
            Designs the section for flexure based on the applied forces.
        check_shear() -> pd.DataFrame:
            Performs a shear check on the section using the applied forces.
        design_shear() -> pd.DataFrame:
            Designs the section for shear based on the applied forces.
        shear_results_detailed(force: Optional[Forces] = None) -> None:
            Provides detailed shear results for a specific force or all forces.
        shear_results_detailed_doc(force: Optional[Forces] = None) -> None:
            Provides detailed shear results documentation for a specific force or all forces.
        flexure_results_detailed(force: Optional[Forces] = None) -> None:
            Provides detailed flexure results for a specific force or all forces.
        flexure_results_detailed_doc(force: Optional[Forces] = None) -> None:
            Provides detailed flexure results documentation for a specific force or all forces.
        results -> None:
            A property that provides beam results for Jupyter Notebook.
        id -> int:
            A read-only property to access the private `_id` of the node.
    """

    _id: int
    _last_id: int = 0  # Class variable to keep track of last assigned ID

    def __init__(self, section: Section, forces: Union[Forces, List[Forces]]) -> None:
        # Initialize the private ID automatically
        Node._last_id += 1  # Increment the class variable for the next ID
        self._id = Node._last_id  # Assign the next available ID

        if not isinstance(section, Section):
            raise TypeError("section must be an instance of Section")

        # Ensure forces is either a single Forces object or a list of Forces
        if isinstance(forces, Forces):
            self.forces = [forces]  # Wrap single Forces object in a list
        elif isinstance(forces, list) and all(
            isinstance(force, Forces) for force in forces
        ):
            self.forces = forces  # Assign the list directly if valid
        else:
            raise TypeError("forces must be an instance of Forces or a list of Forces")

        self.section = section
        # Set the node reference in the section (beam)
        self.section.node = self  # Associate the Node with the RectangularBeam

    def add_forces(self, forces: Union[Forces, List[Forces]]) -> None:
        if isinstance(forces, Forces):
            self.forces.append(forces)  # Append single Forces object
        elif isinstance(forces, list) and all(
            isinstance(force, Forces) for force in forces
        ):
            self.forces.extend(forces)  # Extend list with multiple Forces objects
        else:
            raise TypeError("forces must be an instance of Forces or a list of Forces")

    def get_forces_list(self) -> List[Forces]:
        """Returns the list of forces applied to this node."""
        return self.forces

    def clear_forces(self) -> None:
        """Remove all forces from the node."""
        self.forces.clear()

    def check_flexure(self) -> pd.DataFrame:
        return self.section.check_flexure(self.forces)

    def design_flexure(self) -> pd.DataFrame:
        return self.section.design_flexure(self.forces)

    def check_shear(self) -> pd.DataFrame:
        return self.section.check_shear(self.forces)

    def design_shear(self) -> pd.DataFrame:
        return self.section.design_shear(self.forces)

    def shear_results_detailed(self, force: Optional[Forces] = None) -> None:
        return self.section.shear_results_detailed(force)

    def shear_results_detailed_doc(self, force: Optional[Forces] = None) -> None:
        return self.section.shear_results_detailed_doc(force)

    def flexure_results_detailed(self, force: Optional[Forces] = None) -> None:
        return self.section.flexure_results_detailed(force)

    def flexure_results_detailed_doc(self, force: Optional[Forces] = None) -> None:
        return self.section.flexure_results_detailed_doc(force)

    def __repr__(self) -> str:
        # Get the section label (assuming Section has a `_id` attribute)
        section_label = f"Section label: {self.section.label}"

        # Get the list of forces applied
        forces_list = [
            str(force) for force in self.forces
        ]  # Assuming Forces has a `__repr__` or `__str__` method
        forces_str = (
            "\n  - " + "\n  - ".join(forces_list)
            if forces_list
            else "No forces applied"
        )

        # Combine section label and forces into a single string
        return f"Node ID: {self.id} - {section_label}\nForces Applied:{forces_str}"

    # Beam results for Jupyter Notebook
    @property
    def results(self) -> None:
        return self.section.results

    @property
    def id(self) -> int:
        """Read-only property to access the private _id."""
        return self._id



import pytest
from pint import Quantity
from pint.facets.plain import PlainQuantity
from typing import Generator

from mento.forces import Forces
from mento.units import kN, kNm, kip


# Helper to check if a quantity matches another, allowing for slight float differences
# Option A: Explicitly allow PlainQuantity
def assert_quantity_equal(
    q1: Quantity | PlainQuantity,
    q2: Quantity | PlainQuantity,
    rtol: float = 1e-9,
    atol: float = 1e-12,
) -> None:
    """Asserts two Pint Quantities are numerically and unit-wise equal."""
    assert (
        q1.units == q2.units
    ), f"Units mismatch: {q1.units} (actual) vs {q2.units} (expected)"
    assert (
        pytest.approx(q1.magnitude, rel=rtol, abs=atol) == q2.magnitude
    ), f"Magnitudes mismatch: {q1.magnitude} (actual) vs {q2.magnitude} (expected) with units {q1.units}"


# --- Fixtures ---


@pytest.fixture(autouse=True)
def reset_forces_id_counter() -> Generator[None, None, None]:
    """Fixture to reset the Forces._last_id counter before each test."""
    original_last_id = Forces._last_id
    Forces._last_id = 0
    yield  # Allow the test to run
    Forces._last_id = (
        original_last_id  # Restore if needed, though usually not critical after tests
    )


@pytest.fixture
def default_metric_forces() -> Forces:
    """A Forces instance with default metric values."""
    return Forces()


@pytest.fixture
def custom_metric_forces() -> Forces:
    """A Forces instance with custom metric values."""
    return Forces(label="Custom Metric", N_x=100 * kN, V_z=50 * kN, M_y=200 * kNm)


@pytest.fixture
def custom_imperial_forces() -> Forces:
    """A Forces instance with custom imperial values."""
    # Note: When creating, you can use any compatible units. The class will convert for display.
    return Forces(
        label="Custom Imperial",
        N_x=20 * kip,
        V_z=10 * kip,
        M_y=150 * kNm,
        unit_system="imperial",
    )


# --- Tests for Initialization (`__init__`) ---


def test_forces_default_initialization(default_metric_forces: Forces) -> None:
    """Test Forces initialization with default values."""
    assert default_metric_forces.label is None
    assert_quantity_equal(default_metric_forces._N_x, 0 * kN)
    assert_quantity_equal(default_metric_forces._V_z, 0 * kN)
    assert_quantity_equal(default_metric_forces._M_y, 0 * kNm)
    assert default_metric_forces.unit_system == "metric"
    assert default_metric_forces.id == 1  # First instance after reset


def test_forces_custom_initialization_metric(custom_metric_forces: Forces) -> None:
    """Test Forces initialization with custom metric values."""
    assert custom_metric_forces.label == "Custom Metric"
    assert_quantity_equal(custom_metric_forces._N_x, 100 * kN)
    assert_quantity_equal(custom_metric_forces._V_z, 50 * kN)
    assert_quantity_equal(custom_metric_forces._M_y, 200 * kNm)
    assert custom_metric_forces.unit_system == "metric"
    assert (
        custom_metric_forces.id == 1
    )  # First instance for this test (due to reset fixture)


def test_forces_custom_initialization_imperial(custom_imperial_forces: Forces) -> None:
    """Test Forces initialization with custom imperial values."""
    assert custom_imperial_forces.label == "Custom Imperial"
    assert_quantity_equal(custom_imperial_forces._N_x, 20 * kip)
    assert_quantity_equal(custom_imperial_forces._V_z, 10 * kip)
    # 150 kNm to ft*kip: 150 kN*m * (0.224809 kip/kN) * (3.28084 ft/m) = 110.63 ft*kip approx
    assert_quantity_equal(
        custom_imperial_forces._M_y, 150 * kNm
    )  # Stored internally as kNm
    assert custom_imperial_forces.unit_system == "imperial"
    assert custom_imperial_forces.id == 1


def test_forces_id_incrementing() -> None:
    """Test that the _id increments correctly for multiple instances."""
    f1 = Forces()
    f2 = Forces()
    f3 = Forces()
    assert f1.id == 1
    assert f2.id == 2
    assert f3.id == 3
    assert Forces._last_id == 3


def test_forces_id_is_read_only(default_metric_forces: Forces) -> None:
    """Test that the 'id' property is read-only."""
    with pytest.raises(
        AttributeError, match="property 'id' of 'Forces' object has no setter"
    ):
        default_metric_forces.id = 99  # type: ignore [misc]


# --- Tests for Properties (`N_x`, `V_z`, `M_y`) ---


def test_N_x_property_metric(custom_metric_forces: Forces) -> None:
    """Test N_x property returns correct value in metric."""
    assert_quantity_equal(custom_metric_forces.N_x, 100 * kN)


def test_V_z_property_metric(custom_metric_forces: Forces) -> None:
    """Test V_z property returns correct value in metric."""
    assert_quantity_equal(custom_metric_forces.V_z, 50 * kN)


def test_M_y_property_metric(custom_metric_forces: Forces) -> None:
    """Test M_y property returns correct value in metric."""
    assert_quantity_equal(custom_metric_forces.M_y, 200 * kNm)


def test_N_x_property_imperial(custom_imperial_forces: Forces) -> None:
    """Test N_x property returns correct value in imperial."""
    # 20 kip (already kip)
    assert_quantity_equal(custom_imperial_forces.N_x, 20 * kip)


def test_V_z_property_imperial(custom_imperial_forces: Forces) -> None:
    """Test V_z property returns correct value in imperial."""
    # 10 kip (already kip)
    assert_quantity_equal(custom_imperial_forces.V_z, 10 * kip)


def test_M_y_property_imperial(custom_imperial_forces: Forces) -> None:
    """Test M_y property returns correct value in imperial."""
    # 150 kNm converted to ft*kip (110.63 ft*kip approx)
    assert_quantity_equal(custom_imperial_forces.M_y, (150 * kNm).to("ft*kip"))


# --- Tests for `get_forces()` method ---


def test_get_forces_metric(custom_metric_forces: Forces) -> None:
    """Test get_forces returns correct dictionary in metric."""
    forces_dict = custom_metric_forces.get_forces()
    assert_quantity_equal(forces_dict["N_x"], 100 * kN)
    assert_quantity_equal(forces_dict["V_z"], 50 * kN)
    assert_quantity_equal(forces_dict["M_y"], 200 * kNm)


def test_get_forces_imperial(custom_imperial_forces: Forces) -> None:
    """Test get_forces returns correct dictionary in imperial."""
    forces_dict = custom_imperial_forces.get_forces()
    assert_quantity_equal(forces_dict["N_x"], 20 * kip)
    assert_quantity_equal(forces_dict["V_z"], 10 * kip)
    assert_quantity_equal(forces_dict["M_y"], (150 * kNm).to("ft*kip"))


# --- Tests for `set_forces()` method ---


def test_set_forces_updates_values(default_metric_forces: Forces) -> None:
    """Test set_forces updates the internal force values."""
    default_metric_forces.set_forces(N_x=20 * kN, V_z=5 * kN, M_y=30 * kNm)
    assert_quantity_equal(default_metric_forces._N_x, 20 * kN)
    assert_quantity_equal(default_metric_forces._V_z, 5 * kN)
    assert_quantity_equal(default_metric_forces._M_y, 30 * kNm)


def test_set_forces_with_default_values(default_metric_forces: Forces) -> None:
    """Test set_forces with some default values to ensure they remain unchanged."""
    default_metric_forces.set_forces(N_x=20 * kN)  # Only set N_x
    assert_quantity_equal(default_metric_forces._N_x, 20 * kN)
    assert_quantity_equal(default_metric_forces._V_z, 0 * kN)  # Should remain default
    assert_quantity_equal(default_metric_forces._M_y, 0 * kNm)  # Should remain default


# --- Tests for `compare_to()` method ---


def test_compare_to_V_z_true(custom_metric_forces: Forces) -> None:
    """Test compare_to for V_z when self > other."""
    other_forces = Forces(V_z=25 * kN)
    assert custom_metric_forces.compare_to(other_forces, by="V_z") is True


def test_compare_to_V_z_false(custom_metric_forces: Forces) -> None:
    """Test compare_to for V_z when self <= other."""
    other_forces_equal = Forces(V_z=50 * kN)
    other_forces_greater = Forces(V_z=75 * kN)
    assert custom_metric_forces.compare_to(other_forces_equal, by="V_z") is False
    assert custom_metric_forces.compare_to(other_forces_greater, by="V_z") is False


def test_compare_to_N_x(custom_metric_forces: Forces) -> None:
    """Test compare_to for N_x."""
    other_forces = Forces(N_x=50 * kN)
    assert custom_metric_forces.compare_to(other_forces, by="N_x") is True
    other_forces_larger = Forces(N_x=150 * kN)
    assert custom_metric_forces.compare_to(other_forces_larger, by="N_x") is False


def test_compare_to_M_y(custom_metric_forces: Forces) -> None:
    """Test compare_to for M_y."""
    other_forces = Forces(M_y=150 * kNm)
    assert custom_metric_forces.compare_to(other_forces, by="M_y") is True
    other_forces_larger = Forces(M_y=250 * kNm)
    assert custom_metric_forces.compare_to(other_forces_larger, by="M_y") is False


def test_compare_to_different_units(custom_metric_forces: Forces) -> None:
    """Test compare_to with Forces objects having different internal units."""
    # custom_metric_forces.V_z is 50 kN
    other_forces_imperial = Forces(V_z=10 * kip)  # 10 kip = 44.48 kN
    assert (
        custom_metric_forces.compare_to(other_forces_imperial, by="V_z") is True
    )  # 50 kN > 44.48 kN


def test_compare_to_invalid_attribute(custom_metric_forces: Forces) -> None:
    """Test compare_to raises ValueError for invalid attribute."""
    with pytest.raises(
        ValueError, match="Comparison attribute must be one of 'N_x', 'V_z', or 'M_y'"
    ):
        custom_metric_forces.compare_to(Forces(), by="invalid_attr")


# --- Tests for `__str__()` method ---


def test_str_representation_metric(custom_metric_forces: Forces) -> None:
    """Test __str__ representation for metric forces."""
    s = str(custom_metric_forces)
    # The ID will depend on the order of tests due to the fixture setup, so we only check format
    assert (
        f"Force ID: {custom_metric_forces.id}, Label: Custom Metric, N_x: 100.00 kN, V_z: 50.00 kN, M_y: 200.00 kN·m"
        in s
    )


def test_str_representation_imperial(custom_imperial_forces: Forces) -> None:
    """Test __str__ representation for imperial forces."""
    s = str(custom_imperial_forces)
    assert (
        f"Force ID: {custom_imperial_forces.id}, Label: Custom Imperial, N_x: 20.00 kip, V_z: 10.00 kip, M_y: {custom_imperial_forces.M_y:.2f~P}"
        in s
    )
    # M_y needs careful checking due to potential floating point and pint's default format.
    # The {custom_imperial_forces.M_y:.1f~P} format specifier ensures it's formatted as pint would do for `str()`
    # and matches the expected output including units with unicode mu for 'micro'.



import pytest
import pandas as pd
from docx import Document
from typing import List, Any
from docx.shared import Cm
from mento.results import Formatter, TablePrinter, DocumentBuilder


@pytest.fixture
def formatter() -> Formatter:
    return Formatter()


@pytest.fixture
def table_printer() -> TablePrinter:
    return TablePrinter()


@pytest.fixture
def document_builder() -> DocumentBuilder:
    return DocumentBuilder(title="Test Document")


# Tests for Formatter class
def test_formatter_DCR(formatter: Formatter) -> None:
    assert formatter.DCR(0.85) == "$\\color{#439b00}{\\text{DCR}=0.85}$"
    assert formatter.DCR(0.95) == "$\\color{#efc200}{\\text{DCR}=0.95}$"
    assert formatter.DCR(1.2) == "$\\color{#d43e36}{\\text{DCR}=1.2}$"


def test_formatter_DCR_value(formatter: Formatter) -> None:
    assert formatter.DCR_value(0.85) == "$\\color{#439b00}{0.85}$"
    assert formatter.DCR_value(0.95) == "$\\color{#efc200}{0.95}$"
    assert formatter.DCR_value(1.2) == "$\\color{#d43e36}{1.2}$"


def test_formatter_is_lower(formatter: Formatter) -> None:
    assert formatter.is_lower(0.8, 1.0) == "$\\color{#439b00}{\\, \\checkmark}$"
    assert formatter.is_lower(1.2, 1.0) == "$\\color{#d43e36}{\\, \\times}$"


def test_formatter_is_greater(formatter: Formatter) -> None:
    assert formatter.is_greater(1.2, 1.0) == "$\\color{#439b00}{\\, \\checkmark}$"
    assert formatter.is_greater(0.8, 1.0) == "$\\color{#d43e36}{\\, \\times}$"


def test_formatter_color_DCR_df(formatter: Formatter) -> None:
    df = pd.DataFrame({"DCR": [0.85, 0.95, 1.2]})
    styled_df = formatter.color_DCR_df(df, ["DCR"])
    assert (
        styled_df is not None
    )  # Simply check if the styled DataFrame is generated without errors


# Tests for TablePrinter class
def test_table_printer_data_output(table_printer: TablePrinter, capsys: Any) -> None:
    data = {
        "Column1": ["Value1", "Value2"],
        "Column2": ["a", "b"],
        "Column3": [5.5, 6.7],
        "Column4": ["cm", "mm"],
    }
    headers = ["Column1", "Column2", "Column3", "Column4"]
    table_printer.print_table_data(data, headers)
    captured = capsys.readouterr()
    assert "Value1" in captured.out
    assert "Value2" in captured.out
    assert "5.5" in captured.out
    assert "6.7" in captured.out


# Tests for DocumentBuilder class
def test_document_builder_initialization(document_builder: DocumentBuilder) -> None:
    assert document_builder.title == "Test Document"
    assert isinstance(document_builder.doc, type(Document()))


def test_document_builder_heading(document_builder: DocumentBuilder) -> None:
    document_builder.add_heading("Heading Level 1", level=1)
    assert len(document_builder.doc.paragraphs) > 0
    assert document_builder.doc.paragraphs[0].text == "Heading Level 1"


def test_document_builder_table(document_builder: DocumentBuilder) -> None:
    df = pd.DataFrame({"Name": ["Alice", "Bob"], "Age": [25, 30]})
    column_widths: List[Cm] = [Cm(5), Cm(2)]
    document_builder.add_table(df, column_widths)
    assert len(document_builder.doc.tables) > 0


def test_document_builder_save(
    document_builder: DocumentBuilder, tmp_path: Any
) -> None:
    filename = tmp_path / "test_document.docx"
    document_builder.save(str(filename))
    assert filename.exists()


if __name__ == "__main__":
    pytest.main()



import pytest
import numpy as np

from mento.node import Node
from mento.beam import RectangularBeam
from mento.material import Concrete_ACI_318_19, SteelBar, Concrete_EN_1992_2004
from mento.units import psi, kip, inch, ksi, mm, kN, cm, MPa, ft, kNm
from mento.forces import Forces
from mento.codes.ACI_318_19_beam import (
    _determine_nominal_moment_simple_reinf_ACI_318_19,
    _determine_nominal_moment_double_reinf_ACI_318_19,
)


@pytest.fixture()
def beam_example_imperial() -> RectangularBeam:
    concrete = Concrete_ACI_318_19(name="C4", f_c=4000 * psi)
    steelBar = SteelBar(name="ADN 420", f_y=60 * ksi)
    section = RectangularBeam(
        label="V101",
        concrete=concrete,
        steel_bar=steelBar,
        width=10 * inch,
        height=16 * inch,
        c_c=1.5 * inch,
    )
    return section


@pytest.fixture()
def beam_example_EN_1992_2004() -> RectangularBeam:
    concrete = Concrete_EN_1992_2004(name="C25", f_c=25 * MPa)
    steelBar = SteelBar(name="B500S", f_y=500 * MPa)
    section = RectangularBeam(
        label="101",
        concrete=concrete,
        steel_bar=steelBar,
        width=20 * cm,
        height=60 * cm,
        c_c=2.6 * cm,
    )
    return section


def beam_example_EN_1992_2004_2() -> RectangularBeam:
    concrete = Concrete_EN_1992_2004(name="C25", f_c=25 * MPa)
    steelBar = SteelBar(name="B500S", f_y=500 * MPa)
    section = RectangularBeam(
        label="101",
        concrete=concrete,
        steel_bar=steelBar,
        width=30 * cm,
        height=50 * cm,
        c_c=3.8 * cm,
    )
    return section


def beam_example_EN_1992_2004_3() -> RectangularBeam:
    concrete = Concrete_EN_1992_2004(name="C30", f_c=30 * MPa)
    steelBar = SteelBar(name="B500S", f_y=500 * MPa)
    section = RectangularBeam(
        label="101",
        concrete=concrete,
        steel_bar=steelBar,
        width=30 * cm,
        height=50 * cm,
        c_c=3 * cm,
    )
    return section


def test_shear_check_EN_1992_2004_rebar_1(
    beam_example_EN_1992_2004: RectangularBeam,
) -> None:
    f = Forces(V_z=100 * kN)
    beam_example_EN_1992_2004.set_transverse_rebar(
        n_stirrups=1, d_b=6 * mm, s_l=25 * cm
    )
    beam_example_EN_1992_2004.set_longitudinal_rebar_bot(n1=4, d_b1=16 * mm)
    node = Node(section=beam_example_EN_1992_2004, forces=f)
    results = node.check_shear()

    # Compare dictionaries with a tolerance for floating-point values, in m
    assert results.iloc[1]["Av,min"] == pytest.approx(1.6, rel=1e-3)
    assert results.iloc[1]["Av,req"] == pytest.approx(1.825, rel=1e-3)
    assert results.iloc[1]["Av"] == pytest.approx(2.262, rel=1e-3)
    assert results.iloc[1]["VEd,1"] == pytest.approx(100, rel=1e-3)
    assert results.iloc[1]["VEd,2"] == pytest.approx(100, rel=1e-3)
    assert results.iloc[1]["VRd,c"] == pytest.approx(0, rel=1e-3)
    assert results.iloc[1]["VRd,s"] == pytest.approx(123.924, rel=1e-3)
    assert results.iloc[1]["VRd"] == pytest.approx(123.924, rel=1e-3)
    assert results.iloc[1]["VRd,max"] == pytest.approx(312.811, rel=1e-3)
    assert results.iloc[1]["DCR"] == pytest.approx(0.8069, rel=1e-3)

    # Assert non-numeric values directly
    assert results.iloc[1]["VEd,1<VRd,max"] is True
    assert results.iloc[1]["VEd,2<VRd"] is True


def test_shear_check_EN_1992_2004_rebar_2(
    beam_example_EN_1992_2004: RectangularBeam,
) -> None:
    f = Forces(V_z=350 * kN)
    beam_example_EN_1992_2004.set_transverse_rebar(
        n_stirrups=1, d_b=6 * mm, s_l=25 * cm
    )
    beam_example_EN_1992_2004.set_longitudinal_rebar_bot(n1=4, d_b1=16 * mm)
    node = Node(section=beam_example_EN_1992_2004, forces=f)
    results = node.check_shear()

    # Compare dictionaries with a tolerance for floating-point values, in m
    assert results.iloc[1]["Av,min"] == pytest.approx(1.6, rel=1e-3)
    assert results.iloc[1]["Av,req"] == pytest.approx(7.533, rel=1e-3)
    assert results.iloc[1]["Av"] == pytest.approx(2.262, rel=1e-3)
    assert results.iloc[1]["VEd,1"] == pytest.approx(350, rel=1e-3)
    assert results.iloc[1]["VEd,2"] == pytest.approx(350, rel=1e-3)
    assert results.iloc[1]["VRd,c"] == pytest.approx(0, rel=1e-3)
    assert results.iloc[1]["VRd,s"] == pytest.approx(105.099, rel=1e-3)
    assert results.iloc[1]["VRd"] == pytest.approx(105.099, rel=1e-3)
    assert results.iloc[1]["VRd,max"] == pytest.approx(350, rel=1e-3)
    assert results.iloc[1]["DCR"] == pytest.approx(3.33, rel=1e-3)

    # Assert non-numeric values directly
    assert results.iloc[1]["VEd,1<VRd,max"] is True
    assert results.iloc[1]["VEd,2<VRd"] is False


def test_shear_check_EN_1992_2004_rebar_3(
    beam_example_EN_1992_2004: RectangularBeam,
) -> None:
    f = Forces(V_z=500 * kN)
    beam_example_EN_1992_2004.set_transverse_rebar(
        n_stirrups=1, d_b=6 * mm, s_l=25 * cm
    )
    beam_example_EN_1992_2004.set_longitudinal_rebar_bot(n1=4, d_b1=16 * mm)
    node = Node(section=beam_example_EN_1992_2004, forces=f)
    results = node.check_shear()

    # Compare dictionaries with a tolerance for floating-point values, in m
    assert results.iloc[1]["Av,min"] == pytest.approx(1.6, rel=1e-3)
    assert results.iloc[1]["Av,req"] == pytest.approx(22.817, rel=1e-3)
    assert results.iloc[1]["Av"] == pytest.approx(2.26, rel=1e-3)
    assert results.iloc[1]["VEd,1"] == pytest.approx(500, rel=1e-3)
    assert results.iloc[1]["VEd,2"] == pytest.approx(500, rel=1e-3)
    assert results.iloc[1]["VRd,c"] == pytest.approx(0, rel=1e-3)
    assert results.iloc[1]["VRd,s"] == pytest.approx(49.566, rel=1e-3)
    assert results.iloc[1]["VRd"] == pytest.approx(49.566, rel=1e-3)
    assert results.iloc[1]["VRd,max"] == pytest.approx(453.6, rel=1e-3)
    assert results.iloc[1]["DCR"] == pytest.approx(10.088, rel=1e-3)

    # Assert non-numeric values directly
    assert results.iloc[1]["VEd,1<VRd,max"] is False
    assert results.iloc[1]["VEd,2<VRd"] is False


def test_shear_check_EN_1992_2004_no_rebar_1(
    beam_example_EN_1992_2004: RectangularBeam,
) -> None:
    f = Forces(V_z=30 * kN)
    beam_example_EN_1992_2004.set_longitudinal_rebar_bot(n1=4, d_b1=16 * mm)
    node = Node(section=beam_example_EN_1992_2004, forces=f)
    results = node.check_shear()

    # Compare dictionaries with a tolerance for floating-point values, in m
    assert results.iloc[1]["Av,min"] == pytest.approx(1.6, rel=1e-3)
    assert results.iloc[1]["Av,req"] == pytest.approx(1.6, rel=1e-3)
    assert results.iloc[1]["Av"] == pytest.approx(0, rel=1e-3)
    assert results.iloc[1]["VEd,1"] == pytest.approx(30, rel=1e-3)
    assert results.iloc[1]["VEd,2"] == pytest.approx(30, rel=1e-3)
    assert results.iloc[1]["VRd,c"] == pytest.approx(56.126, rel=1e-3)
    assert results.iloc[1]["VRd,s"] == pytest.approx(0, rel=1e-3)
    assert results.iloc[1]["VRd"] == pytest.approx(56.126, rel=1e-3)
    assert results.iloc[1]["VRd,max"] == pytest.approx(56.126, rel=1e-3)
    assert results.iloc[1]["DCR"] == pytest.approx(0.5345, rel=1e-3)

    # Assert non-numeric values directly
    assert results.iloc[1]["VEd,1<VRd,max"] is True
    assert results.iloc[1]["VEd,2<VRd"] is True


def test_shear_check_EN_1992_2004_no_rebar_2(
    beam_example_EN_1992_2004: RectangularBeam,
) -> None:
    f = Forces(V_z=30 * kN)
    node = Node(section=beam_example_EN_1992_2004, forces=f)
    results = node.check_shear()

    # Compare dictionaries with a tolerance for floating-point values, in m
    assert results.iloc[1]["Av,min"] == pytest.approx(1.6, rel=1e-3)
    assert results.iloc[1]["Av,req"] == pytest.approx(1.6, rel=1e-3)
    assert results.iloc[1]["Av"] == pytest.approx(0, rel=1e-3)
    assert results.iloc[1]["VEd,1"] == pytest.approx(30, rel=1e-3)
    assert results.iloc[1]["VEd,2"] == pytest.approx(30, rel=1e-3)
    assert results.iloc[1]["VRd,c"] == pytest.approx(39.681, rel=1e-3)
    assert results.iloc[1]["VRd,s"] == pytest.approx(0, rel=1e-3)
    assert results.iloc[1]["VRd"] == pytest.approx(39.681, rel=1e-3)
    assert results.iloc[1]["VRd,max"] == pytest.approx(39.681, rel=1e-3)
    assert results.iloc[1]["DCR"] == pytest.approx(0.756, rel=1e-3)

    # Assert non-numeric values directly
    assert results.iloc[1]["VEd,1<VRd,max"] is True
    assert results.iloc[1]["VEd,2<VRd"] is True


def test_shear_check_EN_1992_2004_no_rebar_3(
    beam_example_EN_1992_2004: RectangularBeam,
) -> None:
    f = Forces(N_x=50 * kN, V_z=30 * kN)
    beam_example_EN_1992_2004.set_longitudinal_rebar_bot(n1=4, d_b1=16 * mm)
    node = Node(section=beam_example_EN_1992_2004, forces=f)
    results = node.check_shear()

    # Compare dictionaries with a tolerance for floating-point values, in m
    assert results.iloc[1]["Av,min"] == pytest.approx(1.6, rel=1e-3)
    assert results.iloc[1]["Av,req"] == pytest.approx(1.6, rel=1e-3)
    assert results.iloc[1]["Av"] == pytest.approx(0, rel=1e-3)
    assert results.iloc[1]["VEd,1"] == pytest.approx(30, rel=1e-3)
    assert results.iloc[1]["VEd,2"] == pytest.approx(30, rel=1e-3)
    assert results.iloc[1]["VRd,c"] == pytest.approx(63.101, rel=1e-3)
    assert results.iloc[1]["VRd,s"] == pytest.approx(0, rel=1e-3)
    assert results.iloc[1]["VRd"] == pytest.approx(63.101, rel=1e-3)
    assert results.iloc[1]["VRd,max"] == pytest.approx(63.101, rel=1e-3)
    assert results.iloc[1]["DCR"] == pytest.approx(0.4754, rel=1e-3)

    # Assert non-numeric values directly
    assert results.iloc[1]["VEd,1<VRd,max"] is True
    assert results.iloc[1]["VEd,2<VRd"] is True


def test_shear_design_EN_1992_2004_1(
    beam_example_EN_1992_2004: RectangularBeam,
) -> None:
    f = Forces(N_x=0 * kN, V_z=30 * kN)
    beam_example_EN_1992_2004.set_longitudinal_rebar_bot(n1=4, d_b1=16 * mm)
    node = Node(section=beam_example_EN_1992_2004, forces=f)
    results = node.design_shear()

    # Compare dictionaries with a tolerance for floating-point values, in m
    assert results.iloc[1]["Av,min"] == pytest.approx(1.6, rel=1e-3)
    assert results.iloc[1]["Av,req"] == pytest.approx(1.6, rel=1e-3)
    assert results.iloc[1]["Av"] == pytest.approx(1.616, rel=1e-3)
    assert results.iloc[1]["VEd,1"] == pytest.approx(30, rel=1e-3)
    assert results.iloc[1]["VEd,2"] == pytest.approx(30, rel=1e-3)
    assert results.iloc[1]["VRd,c"] == pytest.approx(0, rel=1e-3)
    assert results.iloc[1]["VRd,s"] == pytest.approx(88.52, rel=1e-3)
    assert results.iloc[1]["VRd"] == pytest.approx(88.52, rel=1e-3)
    assert results.iloc[1]["VRd,max"] == pytest.approx(312.81, rel=1e-3)
    assert results.iloc[1]["DCR"] == pytest.approx(0.339, rel=1e-3)

    # Assert non-numeric values directly
    assert results.iloc[1]["VEd,1<VRd,max"] is True
    assert results.iloc[1]["VEd,2<VRd"] is True


def test_shear_check_ACI_318_19_1(beam_example_imperial: RectangularBeam) -> None:
    f = Forces(V_z=37.727 * kip, N_x=0 * kip)
    beam_example_imperial.set_transverse_rebar(
        n_stirrups=1, d_b=0.5 * inch, s_l=6 * inch
    )
    node = Node(section=beam_example_imperial, forces=f)
    results = node.check_shear()

    # Compare dictionaries with a tolerance for floating-point values, in m
    assert results.iloc[1]["Av,min"] == pytest.approx(2.117, rel=1e-3)
    assert results.iloc[1]["Av,req"] == pytest.approx(10.0623, rel=1e-3)
    assert results.iloc[1]["Av"] == pytest.approx(16.624, rel=1e-3)
    assert results.iloc[1]["ØVc"] == pytest.approx(58.288, rel=1e-3)
    assert results.iloc[1]["ØVs"] == pytest.approx(180.956, rel=1e-3)
    assert results.iloc[1]["ØVn"] == pytest.approx(239.247, rel=1e-3)
    assert results.iloc[1]["ØVmax"] == pytest.approx(291.44, rel=1e-3)
    assert results.iloc[1]["DCR"] == pytest.approx(0.70144, rel=1e-3)

    # Assert non-numeric values directly
    assert results.iloc[1]["Vu<ØVmax"] is True
    assert results.iloc[1]["Vu<ØVn"] is True


def test_shear_check_ACI_318_19_2(beam_example_imperial: RectangularBeam) -> None:
    f = Forces(V_z=37.727 * kip, N_x=20 * kip)
    beam_example_imperial.set_transverse_rebar(
        n_stirrups=1, d_b=0.5 * inch, s_l=6 * inch
    )
    node = Node(section=beam_example_imperial, forces=f)
    results = node.check_shear()

    # Compare dictionaries with a tolerance for floating-point values, in m
    assert results.iloc[1]["Av,min"] == pytest.approx(2.117, rel=1e-3)
    assert results.iloc[1]["Av,req"] == pytest.approx(9.1803, rel=1e-3)
    assert results.iloc[1]["Av"] == pytest.approx(16.624, rel=1e-3)
    assert results.iloc[1]["ØVc"] == pytest.approx(67.888, rel=1e-3)
    assert results.iloc[1]["ØVs"] == pytest.approx(180.959, rel=1e-3)
    assert results.iloc[1]["ØVn"] == pytest.approx(248.847, rel=1e-3)
    assert results.iloc[1]["ØVmax"] == pytest.approx(301.041, rel=1e-3)
    assert results.iloc[1]["DCR"] == pytest.approx(0.6743, rel=1e-3)

    # Assert non-numeric values directly
    assert results.iloc[1]["Vu<ØVmax"] is True
    assert results.iloc[1]["Vu<ØVn"] is True


def test_shear_check_ACI_318_19_no_rebar_1(
    beam_example_imperial: RectangularBeam,
) -> None:
    f = Forces(V_z=8 * kip, N_x=0 * kip)
    beam_example_imperial.set_longitudinal_rebar_bot(n1=2, d_b1=0.625 * inch)
    node = Node(section=beam_example_imperial, forces=f)
    results = node.check_shear()

    # Compare dictionaries with a tolerance for floating-point values, in m
    assert results.iloc[1]["Av,min"] == pytest.approx(2.117, rel=1e-3)
    assert results.iloc[1]["Av,req"] == pytest.approx(2.117, rel=1e-3)
    assert results.iloc[1]["Av"] == pytest.approx(0, rel=1e-3)
    assert beam_example_imperial._d_shear.to("cm").magnitude == pytest.approx(
        35.08, rel=1e-3
    )
    assert results.iloc[1]["ØVc"] == pytest.approx(35.1253, rel=1e-3)
    assert results.iloc[1]["ØVs"] == pytest.approx(0, rel=1e-3)
    assert results.iloc[1]["ØVn"] == pytest.approx(35.125, rel=1e-3)
    assert results.iloc[1]["ØVmax"] == pytest.approx(268.278, rel=1e-3)
    assert results.iloc[1]["DCR"] == pytest.approx(1.013, rel=1e-3)

    # Assert non-numeric values directly
    assert results.iloc[1]["Vu<ØVmax"] is True
    assert results.iloc[1]["Vu<ØVn"] is False


def test_shear_check_ACI_318_19_no_rebar_2(
    beam_example_imperial: RectangularBeam,
) -> None:
    f = Forces(V_z=6 * kip, N_x=0 * kip)
    beam_example_imperial.set_longitudinal_rebar_bot(n1=2, d_b1=0.625 * inch)
    node = Node(section=beam_example_imperial, forces=f)
    results = node.check_shear()

    # Compare dictionaries with a tolerance for floating-point values, in m
    assert results.iloc[1]["Av,min"] == pytest.approx(0, rel=1e-3)
    assert results.iloc[1]["Av,req"] == pytest.approx(0, rel=1e-3)
    assert results.iloc[1]["Av"] == pytest.approx(0, rel=1e-3)
    assert results.iloc[1]["ØVc"] == pytest.approx(58.288, rel=1e-3)
    assert results.iloc[1]["ØVs"] == pytest.approx(0, rel=1e-3)
    assert results.iloc[1]["ØVn"] == pytest.approx(58.288, rel=1e-3)
    assert results.iloc[1]["ØVmax"] == pytest.approx(291.441, rel=1e-3)
    assert results.iloc[1]["DCR"] == pytest.approx(0.4579, rel=1e-3)

    # Assert non-numeric values directly
    assert results.iloc[1]["Vu<ØVmax"] is True
    assert results.iloc[1]["Vu<ØVn"] is True


def test_shear_design_ACI_318_19(beam_example_imperial: RectangularBeam) -> None:
    f = Forces(V_z=37.727 * kip, N_x=0 * kip)
    beam_example_imperial.set_longitudinal_rebar_bot(n1=2, d_b1=0.625 * inch)
    node = Node(section=beam_example_imperial, forces=f)
    results = node.design_shear()

    # Compare dictionaries with a tolerance for floating-point values, in m
    assert results.iloc[1]["Av,min"] == pytest.approx(2.117, rel=1e-3)
    assert results.iloc[1]["Av,req"] == pytest.approx(10.06, rel=1e-3)
    assert results.iloc[1]["Av"] == pytest.approx(11.22, rel=1e-3)
    assert results.iloc[1]["ØVc"] == pytest.approx(58.29, rel=1e-3)
    assert results.iloc[1]["ØVs"] == pytest.approx(122.15, rel=1e-3)
    assert results.iloc[1]["ØVn"] == pytest.approx(180.44, rel=1e-3)
    assert results.iloc[1]["ØVmax"] == pytest.approx(291.44, rel=1e-3)
    assert results.iloc[1]["DCR"] == pytest.approx(0.93, rel=1e-3)

    # Assert non-numeric values directly
    assert results.iloc[1]["Vu<ØVmax"] is True
    assert results.iloc[1]["Vu<ØVn"] is True


# # ------- FLEXURE TEST --------------


def test_flexure_check_EN_1992_2004_1(
    beam_example_EN_1992_2004: RectangularBeam,
) -> None:
    # Example from Calcpad Validation
    f = Forces(M_y=150 * kNm)
    beam_example_EN_1992_2004.set_longitudinal_rebar_bot(n1=4, d_b1=16 * mm)
    node = Node(section=beam_example_EN_1992_2004, forces=f)
    results = node.check_flexure()

    assert results.iloc[0]["Position"] == "Bottom"
    assert results.iloc[0]["As,min"].to(cm**2).magnitude == pytest.approx(
        1.49, rel=1e-2
    )
    assert results.iloc[0]["As,req bot"].to(cm**2).magnitude == pytest.approx(
        6.68, rel=1e-3
    )
    assert results.iloc[0]["As,req top"].to(cm**2).magnitude == pytest.approx(
        0, rel=1e-3
    )
    assert results.iloc[0]["As"].to(inch**2).magnitude == pytest.approx(
        8.04, rel=1e-2
    )
    assert results.iloc[0]["MRd"].to(kip * ft).magnitude == pytest.approx(
        180.47, rel=1e-5
    )
    assert results.iloc[0]["DCR"] == pytest.approx(0.831, rel=1e-5)


def test_flexure_check_EN_1992_2004_2(
    beam_example_EN_1992_2004: RectangularBeam,
) -> None:
    # Example from Calcpad Validation
    f = Forces(M_y=450 * kNm)
    beam_example_EN_1992_2004.set_longitudinal_rebar_bot(n1=5, d_b1=25 * mm)
    node = Node(section=beam_example_EN_1992_2004, forces=f)
    results = node.check_flexure()

    assert results.iloc[0]["Position"] == "Bottom"
    assert results.iloc[0]["As,min"].to(cm**2).magnitude == pytest.approx(
        1.48, rel=1e-3
    )
    assert results.iloc[0]["As,req bot"].to(cm**2).magnitude == pytest.approx(
        21.07, rel=1e-3
    )
    assert results.iloc[0]["As,req top"].to(cm**2).magnitude == pytest.approx(
        8.64, rel=1e-3
    )
    assert results.iloc[0]["As"].to(inch**2).magnitude == pytest.approx(
        24.54, rel=1e-3
    )
    assert results.iloc[0]["MRd"].to(kip * ft).magnitude == pytest.approx(
        505.6, rel=1e-3
    )
    assert results.iloc[0]["DCR"] == pytest.approx(0.89, rel=1e-5)


def test_flexure_check_EN_1992_2004_3(
    beam_example_EN_1992_2004_2: RectangularBeam,
) -> None:
    # Example from Designers Guide to EN 1992-1-1:2004, Example 5.1, Page 66
    f = Forces(M_y=120 * kNm)
    node = Node(section=beam_example_EN_1992_2004_2, forces=f)
    results = node.check_flexure()

    assert results.iloc[0]["Position"] == "Bottom"
    assert results.iloc[0]["As,min"].to(cm**2).magnitude == pytest.approx(
        1.8, rel=1e-2
    )
    assert results.iloc[0]["As,req bot"].to(cm**2).magnitude == pytest.approx(
        6.63, rel=1e-3
    )
    assert results.iloc[0]["As,req top"].to(cm**2).magnitude == pytest.approx(
        0, rel=1e-3
    )


def test_flexure_check_EN_1992_2004_4(
    beam_example_EN_1992_2004_3: RectangularBeam,
) -> None:
    # Example from Lecture-3-Bending-and-Shear-in-Beams-Concrete Centre - Page 14
    f = Forces(M_y=370 * kNm)
    beam_example_EN_1992_2004_3.set_longitudinal_rebar_bot(n1=2, d_b1=32 * mm)
    beam_example_EN_1992_2004_3.set_longitudinal_rebar_top(n1=2, d_b1=20 * mm)
    beam_example_EN_1992_2004_3.set_transverse_rebar(
        n_stirrups=1, d_b=10 * mm, s_l=20 * cm
    )
    node = Node(section=beam_example_EN_1992_2004_3, forces=f)
    results = node.check_flexure()

    assert results.iloc[0]["Position"] == "Bottom"
    assert results.iloc[0]["As,min"].to(cm**2).magnitude == pytest.approx(
        2.01, rel=1e-3
    )
    assert results.iloc[0]["As,req bot"].to(cm**2).magnitude == pytest.approx(
        23.07, rel=1e-3  # 23.07 in Example
    )
    assert results.iloc[0]["As,req top"].to(cm**2).magnitude == pytest.approx(
        4.27, rel=1e-3  # 4.27 in Example
    )


def test_flexural_beam_determine_nominal_moment_simple_reinf_ACI_318_19() -> None:
    """
    Test para la funcion que determina el momento nominal de una seccion armada solo con refuerzo de traccion
    ver el ejemplo en:
    Example from https://www.google.com/search?q=ACI+318+19+nominal+moment+of+section&oq=ACI+318+19+nominal+moment+of+section&gs_lcrp=EgZjaHJvbWUyBggAEEUYOTIHCAEQIRigAdIBCTIyNzkzajBqN6gCALACAA&sourceid=chrome&ie=UTF-8#fpstate=ive&vld=cid:dab5f5e7,vid:m4H0QbGDYIg,st:0
    """
    concrete = Concrete_ACI_318_19(name="C6", f_c=6000 * psi)
    steelBar = SteelBar(name="G80", f_y=80000 * psi)
    beam = RectangularBeam(
        label="B20x30",
        concrete=concrete,
        steel_bar=steelBar,
        width=20 * inch,
        height=30 * inch,
        c_c=1 * inch,
    )
    A_s = 10.92 * inch**2
    d = 27 * inch
    result = _determine_nominal_moment_simple_reinf_ACI_318_19(beam, A_s, d)

    # Compare dictionaries with a tolerance for floating-point values, in inch**2
    # Result: 1653.84kip.ft
    assert result.to(kip * ft).magnitude == pytest.approx(1653.84, rel=1e-2)


def test_flexural_beam_determine_nominal_moment_double_reinf_ACI_318_19() -> None:
    """
    Test para la funcion que determina el momento nominal de una seccion doblemente armada
    Example from https://www.youtube.com/watch?v=7BOuV1gCcgc
    """
    concrete = Concrete_ACI_318_19(name="C4", f_c=4000 * psi)
    steelBar = SteelBar(name="G60", f_y=60000 * psi)
    beam = RectangularBeam(
        label="B14x27",
        concrete=concrete,
        steel_bar=steelBar,
        width=14 * inch,
        height=27 * inch,
        c_c=1 * inch,
    )
    A_s = 6 * inch**2
    d = 24 * inch
    d_prime = 2.5 * inch
    A_s_prime = 1.8 * inch**2
    result = _determine_nominal_moment_double_reinf_ACI_318_19(
        beam, A_s, d, d_prime, A_s_prime
    )
    assert result.to(kip * ft).magnitude == pytest.approx(639.12, rel=1e-2)


# BEAM FROM CALPCAD "ACI 318-19 Beam Flexure 02-TEST_1"
@pytest.fixture()
def beam_example_flexure_ACI() -> RectangularBeam:
    concrete = Concrete_ACI_318_19(name="fc 4000", f_c=4000 * psi)
    steelBar = SteelBar(name="fy 60000", f_y=60 * ksi)
    section = RectangularBeam(
        label="B-12x24",
        concrete=concrete,
        steel_bar=steelBar,
        width=12 * inch,
        height=24 * inch,
        c_c=1.5 * inch,
    )
    return section


def test_check_flexure_ACI_318_19_1(beam_example_flexure_ACI: RectangularBeam) -> None:
    # Testing the check of the reinforced beam with a large moment that requires compression reinforcement, the moment being positive
    # # See calcpad: ACI 318-19 Beam Flexure 02-TEST_1.cpd
    f = Forces(label="Test_01", M_y=400 * kip * ft)
    beam_example_flexure_ACI.set_longitudinal_rebar_bot(
        n1=2, d_b1=1.41 * inch, n3=2, d_b3=1.27 * inch
    )
    beam_example_flexure_ACI.set_longitudinal_rebar_top(n1=2, d_b1=0.75 * inch)
    node = Node(section=beam_example_flexure_ACI, forces=f)
    results = node.check_flexure()

    assert results.iloc[0]["Section Label"] == "B-12x24"
    assert results.iloc[0]["Load Combo"] == "Test_01"
    assert results.iloc[0]["Position"] == "Bottom"
    assert results.iloc[0]["As,min"].to(cm**2).magnitude == pytest.approx(
        5.26, rel=1e-2
    )
    assert results.iloc[0]["As,req bot"].to(cm**2).magnitude == pytest.approx(
        33.1712108, rel=1e-3
    )
    assert results.iloc[0]["As,req top"].to(cm**2).magnitude == pytest.approx(
        4.9250577, rel=1e-3
    )
    assert results.iloc[0]["As"].to(inch**2).magnitude == pytest.approx(
        5.66, rel=1e-2
    )
    assert results.iloc[0]["Mu"].to(kip * ft).magnitude == pytest.approx(400, rel=1e-5)


def test_check_flexure_ACI_318_19_2(beam_example_flexure_ACI: RectangularBeam) -> None:
    # Testeo el checkeo de la viga armada con un momento grande que requiere armadura de compresión, siendo el momento NEGATIVO.
    # Ver calcpad: ACI 318-19 Beam Flexure 02-TEST_1.cpd
    # Invierto el armado respecto al test anterior para poder usar el mismo calcpad
    f = Forces(label="Test_02", M_y=-400 * kip * ft)
    beam_example_flexure_ACI.set_longitudinal_rebar_bot(n1=2, d_b1=0.75 * inch)
    beam_example_flexure_ACI.set_longitudinal_rebar_top(
        n1=2, d_b1=1.41 * inch, n3=2, d_b3=1.27 * inch
    )
    node = Node(section=beam_example_flexure_ACI, forces=f)
    results = node.check_flexure()

    assert results.iloc[0]["Section Label"] == "B-12x24"
    assert results.iloc[0]["Load Combo"] == "Test_02"
    assert results.iloc[0]["Position"] == "Top"
    assert results.iloc[0]["As,min"].to(cm**2).magnitude == pytest.approx(
        5.26, rel=1e-2
    )
    assert results.iloc[0]["As,req bot"].to(cm**2).magnitude == pytest.approx(
        4.9250577, rel=1e-3
    )
    assert results.iloc[0]["As,req top"].to(cm**2).magnitude == pytest.approx(
        33.1712108, rel=1e-3
    )
    assert results.iloc[0]["As"].to(inch**2).magnitude == pytest.approx(
        5.66, rel=1e-2
    )
    assert results.iloc[0]["Mu"].to(kip * ft).magnitude == pytest.approx(-400, rel=1e-5)


# def test_design_flexure_ACI_318_19_1(beam_example_flexure_ACI: RectangularBeam) -> None:
#     f = Forces(label='Test_01', M_y=400*kip*ft)
#     forces=[f]
#     results = beam_example_flexure_ACI._design_flexure(forces)

#     assert results.iloc[0]['Section Label'] == 'B-12x24'
#     assert results.iloc[0]['Load Combo']  == 'Test_01'
#     assert results.iloc[0]['Position'] == 'Bottom'
#     assert results.iloc[0]['As,min'].to(cm**2).magnitude == pytest.approx(5.26, rel=1e-2)
#     assert results.iloc[0]['As,req bot'].to(cm**2).magnitude == pytest.approx(33.17, rel=1e-3)
#     assert results.iloc[0]['As,req top'].to(cm**2).magnitude == pytest.approx(4.93, rel=1e-2)
#     assert results.iloc[0]['As'].to(inch**2).magnitude == pytest.approx(5.66, rel=1e-2)


def testing_determine_nominal_moment_ACI_318_19(
    beam_example_flexure_ACI: RectangularBeam,
) -> None:
    beam_example_flexure_ACI.set_longitudinal_rebar_bot(
        n1=2,
        d_b1=1.128 * inch,
        n2=1,
        d_b2=1.128 * inch,
        n3=2,
        d_b3=1 * inch,
        n4=1,
        d_b4=1 * inch,
    )
    beam_example_flexure_ACI.set_longitudinal_rebar_top(
        n1=2,
        d_b1=1.128 * inch,
        n2=1,
        d_b2=1.128 * inch,
        n3=2,
        d_b3=1 * inch,
        n4=1,
        d_b4=1 * inch,
    )

    f = Forces(label="Test_01", M_y=400 * kip * ft)
    node_1 = Node(section=beam_example_flexure_ACI, forces=f)
    node_1.check_flexure()
    assert beam_example_flexure_ACI._phi_M_n_bot.to(kNm).magnitude == pytest.approx(
        587.0589108678, rel=1e-2
    )
    assert beam_example_flexure_ACI._phi_M_n_top.to(kNm).magnitude == pytest.approx(
        587.0589108678, rel=1e-2
    )


# This is where pytest will collect the tests and run them
if __name__ == "__main__":
    pytest.main()



import pytest
import math
from pint import Quantity

# Import all classes from your material.py and relevant units
from mento.material import (
    Material,
    Concrete,
    Concrete_ACI_318_19,
    Concrete_CIRSOC_201_25,
    Concrete_EN_1992_2004,
    Steel,
    SteelBar,
    SteelStrand,
)
from mento.units import kg, m, MPa, ksi, GPa, psi, Pa, lb, ft, kPa, mm


# Helper to check if a quantity matches another, allowing for slight float differences
def assert_quantity_equal(
    q1: Quantity, q2: Quantity, rtol: float = 1e-6, atol: float = 1e-9
) -> None:
    """Asserts two Pint Quantities are numerically and unit-wise equal."""
    assert q1.units == q2.units, f"Units mismatch: {q1.units} vs {q2.units}"
    assert (
        pytest.approx(q1.magnitude, rel=rtol, abs=atol) == q2.magnitude
    ), f"Magnitudes mismatch: {q1.magnitude} (actual) vs {q2.magnitude} (expected) with units {q1.units}"


# --- Tests for Material (Base Class) ---


def test_material_initialization() -> None:
    """Test the basic Material class initialization."""
    mat = Material(name="Generic Material")
    assert mat.name == "Generic Material"


# --- Tests for Concrete (Base Concrete Class) ---


def test_concrete_initialization_metric_mpa() -> None:
    """Test Concrete initialization with f_c in MPa (metric)."""
    concrete = Concrete(name="H30", f_c=30 * MPa)
    assert concrete.name == "H30"  # Correctly check the name set during initialization
    assert_quantity_equal(concrete.f_c, 30 * MPa)
    assert concrete.design_code == "ACI 318-19"  # Default
    assert concrete.unit_system == "metric"
    assert_quantity_equal(concrete.density, 2500 * kg / m**3)


def test_concrete_initialization_metric_pa() -> None:
    """Test Concrete initialization with f_c in Pa (metric)."""
    concrete = Concrete(name="H25", f_c=25e6 * Pa)
    assert concrete.unit_system == "metric"
    assert_quantity_equal(concrete.density, 2500 * kg / m**3)
    assert_quantity_equal(concrete.f_c, 25e6 * Pa)  # Should convert to MPa implicitly


def test_concrete_initialization_metric_kpa() -> None:
    """Test Concrete initialization with f_c in kPa (metric)."""
    concrete = Concrete(name="H25", f_c=25000 * kPa)
    assert concrete.unit_system == "metric"
    assert_quantity_equal(concrete.density, 2500 * kg / m**3)
    assert_quantity_equal(concrete.f_c, 25000 * kPa)


def test_concrete_initialization_imperial_psi() -> None:
    """Test Concrete initialization with f_c in psi (imperial)."""
    concrete = Concrete(name="C4", f_c=4000 * psi)
    assert_quantity_equal(concrete.f_c, 4000 * psi)
    assert concrete.unit_system == "imperial"
    assert_quantity_equal(concrete.density, 155 * lb / ft**3)


def test_concrete_initialization_imperial_ksi() -> None:
    """Test Concrete initialization with f_c in ksi (imperial)."""
    concrete = Concrete(name="C4", f_c=5 * ksi)
    assert_quantity_equal(concrete.f_c, 5 * ksi)
    assert concrete.unit_system == "imperial"
    assert_quantity_equal(concrete.density, 155 * lb / ft**3)


def test_concrete_initialization_unsupported_unit() -> None:
    """Test Concrete initialization with an unsupported unit for f_c."""
    with pytest.raises(ValueError, match="Unsupported unit system for f_c"):
        Concrete(name="C25", f_c=25 * m)


def test_concrete_get_properties() -> None:
    """Test get_properties method of Concrete base class."""
    concrete = Concrete(name="C30", f_c=30 * MPa)
    props = concrete.get_properties()
    assert_quantity_equal(props["f_c"], 30 * MPa)
    assert_quantity_equal(props["density"], 2500 * kg / m**3)
    assert (
        "design_code" not in props
    )  # Base get_properties only includes f_c and density


# --- Tests for Concrete_ACI_318_19 ---


@pytest.fixture
def aci_metric_concrete() -> Concrete_ACI_318_19:
    """Fixture for a standard metric ACI concrete."""
    return Concrete_ACI_318_19(f_c=25 * MPa, name="C25_ACI")


@pytest.fixture
def aci_imperial_concrete() -> Concrete_ACI_318_19:
    """Fixture for a standard imperial ACI concrete."""
    return Concrete_ACI_318_19(f_c=4000 * psi, name="4ksi_ACI")


def test_aci_initialization_metric(aci_metric_concrete: Concrete_ACI_318_19) -> None:
    """Test ACI concrete initialization in metric."""
    assert aci_metric_concrete.name == "C25_ACI"
    assert_quantity_equal(aci_metric_concrete.f_c, 25 * MPa)
    assert aci_metric_concrete.design_code == "ACI 318-19"
    assert aci_metric_concrete.unit_system == "metric"
    assert_quantity_equal(aci_metric_concrete.density, 2500 * kg / m**3)

    # Expected values for 25 MPa concrete (metric)
    # E_c = (2500^1.5) * 0.043 * sqrt(25) * MPa = 125000 * 0.043 * 5 * MPa = 26875 MPa
    assert_quantity_equal(aci_metric_concrete.E_c, 26875 * MPa)
    # f_r = 0.625 * sqrt(25) * MPa = 0.625 * 5 * MPa = 3.125 MPa
    assert_quantity_equal(aci_metric_concrete.f_r, 3.125 * MPa)
    # beta_1 for 25 MPa (17 <= 25 <= 28) should be 0.85
    assert pytest.approx(aci_metric_concrete.beta_1) == 0.85
    assert pytest.approx(aci_metric_concrete.lambda_factor) == 1.0  # Normalweight
    assert pytest.approx(aci_metric_concrete.phi_v) == 0.75
    assert pytest.approx(aci_metric_concrete.phi_c) == 0.65
    assert (
        pytest.approx(aci_metric_concrete.phi_y) == 0.90
    )  # Note: property is phi_y but uses _phi_t


def test_aci_initialization_imperial(
    aci_imperial_concrete: Concrete_ACI_318_19,
) -> None:
    """Test ACI concrete initialization in imperial."""
    assert aci_imperial_concrete.name == "4ksi_ACI"
    assert_quantity_equal(aci_imperial_concrete.f_c, 4000 * psi)
    assert aci_imperial_concrete.design_code == "ACI 318-19"
    assert aci_imperial_concrete.unit_system == "imperial"
    assert_quantity_equal(aci_imperial_concrete.density, 155 * lb / ft**3)

    # Expected values for 4000 psi concrete (imperial)
    # E_c = (155^1.5) * 33 * sqrt(4000) * psi = 1930.5 * 33 * 63.245 * psi = 4030635.7 psi ~ 4.03e6 psi
    assert_quantity_equal(
        aci_imperial_concrete.E_c, (155**1.5) * 33 * math.sqrt(4000) * psi
    )
    # f_r = 7.5 * sqrt(4000) * psi = 7.5 * 63.245 * psi = 474.34 psi
    assert_quantity_equal(aci_imperial_concrete.f_r, 7.5 * math.sqrt(4000) * psi)
    # For beta_1, f_c must be in MPa. 4000 psi = 27.579 MPa.
    # 17 <= 27.579 <= 28, so beta_1 should be 0.85
    assert pytest.approx(aci_imperial_concrete.beta_1) == 0.85


# Test __beta_1 calculation ranges
@pytest.mark.parametrize(
    "f_c_value, expected_beta_1",
    [
        (17 * MPa, 0.85),
        (28 * MPa, 0.85),
        (20 * MPa, 0.85),  # In the first range
        (30 * MPa, 0.85 - 0.05 / 7 * (30 - 28)),  # 28 < fc <= 55 range
        (40 * MPa, 0.85 - 0.05 / 7 * (40 - 28)),
        (55 * MPa, 0.85 - 0.05 / 7 * (55 - 28)),
        (60 * MPa, 0.65),  # fc > 55 range
        (70 * MPa, 0.65),
        (10 * MPa, 0.85),  # fc < 17 MPa, falls into the else branch (0.85)
    ],
)
def test_aci_beta_1_calculation(f_c_value: Quantity, expected_beta_1: float) -> None:
    concrete = Concrete_ACI_318_19(name="C4", f_c=f_c_value)
    assert pytest.approx(concrete.beta_1) == expected_beta_1


def test_aci_get_properties(aci_metric_concrete: Concrete_ACI_318_19) -> None:
    """Test get_properties for ACI concrete."""
    props = aci_metric_concrete.get_properties()
    assert_quantity_equal(props["f_c"], aci_metric_concrete.f_c)
    assert_quantity_equal(props["density"], aci_metric_concrete.density)
    assert_quantity_equal(props["E_c"], aci_metric_concrete.E_c)
    assert_quantity_equal(props["f_r"], aci_metric_concrete.f_r)
    assert pytest.approx(props["beta_1"]) == aci_metric_concrete.beta_1
    assert pytest.approx(props["epsilon_c"]) == aci_metric_concrete._epsilon_c
    assert pytest.approx(props["lambda"]) == aci_metric_concrete.lambda_factor
    assert pytest.approx(props["phi_v"]) == aci_metric_concrete.phi_v
    assert pytest.approx(props["phi_c"]) == aci_metric_concrete.phi_c
    assert (
        pytest.approx(props["phi_t"]) == aci_metric_concrete.phi_y
    )  # Check the name mapping


def test_aci_str_representation(aci_metric_concrete: Concrete_ACI_318_19) -> None:
    """Test __str__ method for ACI concrete."""
    s = str(aci_metric_concrete)
    assert "Concrete Properties (C25_ACI):" in s
    assert f"  f_c: {aci_metric_concrete.f_c}" in s
    assert f"  Density: {aci_metric_concrete.density}" in s
    assert f"  E_c: {aci_metric_concrete.E_c}" in s
    assert f"  f_r: {aci_metric_concrete.f_r}" in s
    assert f"  beta_1: {aci_metric_concrete.beta_1:.3f}" in s
    assert f"  epsilon_c: {aci_metric_concrete._epsilon_c:.4f}" in s
    assert f"  λ: {aci_metric_concrete.lambda_factor:.2f}" in s
    assert f"  phi_v: {aci_metric_concrete.phi_v:.2f}" in s
    assert f"  phi_c: {aci_metric_concrete.phi_c:.2f}" in s
    assert (
        f"  phi_t: {aci_metric_concrete.phi_y:.2f}" in s
    )  # Using phi_y as per __str__ output


# --- Tests for Concrete_CIRSOC_201_25 ---


def test_cirsoc_initialization() -> None:
    """Test CIRSOC concrete initialization."""
    concrete = Concrete_CIRSOC_201_25(f_c=30 * MPa, name="H30_CIRSOC")
    assert concrete.name == "H30_CIRSOC"
    assert_quantity_equal(concrete.f_c, 30 * MPa)
    assert concrete.design_code == "CIRSOC 201-25"
    assert concrete.unit_system == "metric"
    assert_quantity_equal(concrete.density, 2500 * kg / m**3)
    # Verify inherited properties are calculated based on ACI logic
    assert_quantity_equal(concrete.E_c, ((2500**1.5) * 0.043 * math.sqrt(30)) * MPa)
    assert pytest.approx(concrete.beta_1) == (
        0.85 - 0.05 / 7 * (30 - 28)
    )  # Should follow ACI beta_1 logic


# --- Tests for Concrete_EN_1992_2004 ---


@pytest.fixture
def en_concrete_c30() -> Concrete_EN_1992_2004:
    """Fixture for EN concrete, C30/37 class (fck=30 MPa)."""
    return Concrete_EN_1992_2004(f_c=30 * MPa, name="C30/37")


@pytest.fixture
def en_concrete_c60() -> Concrete_EN_1992_2004:
    """Fixture for EN concrete, C60/75 class (fck=60 MPa)."""
    return Concrete_EN_1992_2004(f_c=60 * MPa, name="C60/75")


def test_en_initialization_c30(en_concrete_c30: Concrete_EN_1992_2004) -> None:
    """Test EN concrete initialization for C30."""
    assert en_concrete_c30.name == "C30/37"
    assert_quantity_equal(en_concrete_c30.f_c, 30 * MPa)  # This is f_ck
    assert en_concrete_c30.design_code == "EN 1992-2004"
    assert en_concrete_c30.unit_system == "metric"
    assert_quantity_equal(en_concrete_c30.density, 2500 * kg / m**3)

    # Expected values for C30/37 (f_ck=30 MPa)
    assert_quantity_equal(en_concrete_c30.f_ck, 30 * MPa)
    assert_quantity_equal(
        en_concrete_c30.f_cm, 30 * MPa + 8 * MPa
    )  # f_ck + 8 MPa = 38 MPa
    # E_cm = 22000 * (38/10)^0.3 * MPa = 22000 * 3.8^0.3 * MPa = 22000 * 1.488 * MPa = 32736 MPa
    assert_quantity_equal(en_concrete_c30.E_cm, 22000 * (3.8) ** 0.3 * MPa)
    # f_ctm = 0.3 * (30)^(2/3) * MPa = 0.3 * 9.6548 * MPa = 2.89644 MPa
    assert_quantity_equal(en_concrete_c30.f_ctm, 0.3 * (30) ** (2 / 3) * MPa)
    assert pytest.approx(en_concrete_c30.epsilon_cu3) == 0.0035
    assert pytest.approx(en_concrete_c30.gamma_c) == 1.5
    assert pytest.approx(en_concrete_c30.gamma_s) == 1.15
    assert pytest.approx(en_concrete_c30.alpha_cc) == 0.85  # As per _alpha_cc_calc
    assert pytest.approx(en_concrete_c30.Lambda_factor) == 0.8  # For f_ck <= 50 MPa
    assert pytest.approx(en_concrete_c30.Eta_factor) == 1.0  # For f_ck <= 50 MPa


def test_en_initialization_c60(en_concrete_c60: Concrete_EN_1992_2004) -> None:
    """Test EN concrete initialization for C60."""
    assert_quantity_equal(en_concrete_c60.f_c, 60 * MPa)  # This is f_ck
    assert en_concrete_c60.design_code == "EN 1992-2004"

    # Expected values for C60/75 (f_ck=60 MPa)
    assert_quantity_equal(en_concrete_c60.f_ck, 60 * MPa)
    assert_quantity_equal(en_concrete_c60.f_cm, 60 * MPa + 8 * MPa)  # 68 MPa
    assert pytest.approx(en_concrete_c60.Lambda_factor) == (
        0.8 - (60 - 50) / 400
    )  # 0.8 - 10/400 = 0.775
    assert pytest.approx(en_concrete_c60.Eta_factor) == (
        1.0 - (60 - 50) / 200
    )  # 1.0 - 10/200 = 0.95


def test_en_get_properties(en_concrete_c30: Concrete_EN_1992_2004) -> None:
    """Test get_properties for EN concrete."""
    props = en_concrete_c30.get_properties()
    assert_quantity_equal(props["f_ck"], en_concrete_c30.f_ck)
    assert_quantity_equal(props["f_cm"], en_concrete_c30.f_cm)
    assert_quantity_equal(props["E_cm"], en_concrete_c30.E_cm)
    assert_quantity_equal(props["f_ctm"], en_concrete_c30.f_ctm)
    assert pytest.approx(props["epsilon_cu3"]) == en_concrete_c30.epsilon_cu3
    assert pytest.approx(props["gamma_c"]) == en_concrete_c30.gamma_c
    assert pytest.approx(props["gamma_s"]) == en_concrete_c30.gamma_s
    assert pytest.approx(props["alpha_cc"]) == en_concrete_c30.alpha_cc
    assert pytest.approx(props["lambda_factor"]) == en_concrete_c30.Lambda_factor
    assert pytest.approx(props["eta_factor"]) == en_concrete_c30.Eta_factor


def test_en_str_representation(en_concrete_c30: Concrete_EN_1992_2004) -> None:
    """Test __str__ method for EN concrete."""
    s = str(en_concrete_c30)
    assert "Concrete Properties (C30/37):" in s
    assert "Design Code: EN 1992-2004" in s
    assert f"  f_c (Characteristic): {en_concrete_c30.f_ck}" in s
    assert f"  f_cm (Mean Compressive): {en_concrete_c30.f_cm}" in s
    assert f"  f_ctm (Mean Tensile): {en_concrete_c30.f_ctm}" in s
    assert f"  E_cm (Secant Modulus): {en_concrete_c30.E_cm}" in s
    assert f"  Density: {en_concrete_c30.density}" in s
    assert f"  ε_cu3: {en_concrete_c30.epsilon_cu3:.4f}" in s
    assert f"  γ_c: {en_concrete_c30.gamma_c:.2f}" in s
    assert f"  γ_s: {en_concrete_c30.gamma_s:.2f}" in s
    assert f"  α_cc: {en_concrete_c30.alpha_cc:.2f}" in s
    assert f"  λ Factor: {en_concrete_c30.Lambda_factor:.2f}" in s
    assert f"  Eta Factor: {en_concrete_c30.Eta_factor:.2f}" in s


# --- Tests for Steel (Base Class) ---


def test_steel_initialization() -> None:
    """Test the Steel base class initialization."""
    steel = Steel(name="Generic Steel", f_y=400 * MPa)
    assert steel.name == "Generic Steel"
    assert_quantity_equal(steel.f_y, 400 * MPa)
    assert_quantity_equal(steel.density, 7850 * kg / m**3)


def test_steel_initialization_custom_density() -> None:
    """Test Steel initialization with custom density."""
    custom_density = 7500 * kg / m**3
    steel = Steel(name="Light Steel", f_y=350 * MPa, density=custom_density)
    assert_quantity_equal(steel.density, custom_density)


# --- Tests for SteelBar ---


@pytest.fixture
def steel_bar_b500s() -> SteelBar:
    """Fixture for a typical SteelBar (B500S)."""
    return SteelBar(name="B500S", f_y=500 * MPa)


@pytest.fixture
def steel_bar_grade60() -> SteelBar:
    """Fixture for a typical SteelBar (Grade 60)."""
    return SteelBar(name="Grade 60", f_y=60 * ksi)


def test_steel_bar_initialization_metric(steel_bar_b500s: SteelBar) -> None:
    """Test SteelBar initialization in metric."""
    assert steel_bar_b500s.name == "B500S"
    assert_quantity_equal(steel_bar_b500s.f_y, 500 * MPa)
    assert_quantity_equal(steel_bar_b500s.E_s, 200 * GPa)
    assert_quantity_equal(steel_bar_b500s.density, 7850 * kg / m**3)
    # epsilon_y = 500 MPa / 200 GPa = 500e6 Pa / 200e9 Pa = 0.0025 (dimensionless)
    assert_quantity_equal(
        steel_bar_b500s.epsilon_y, 0.0025 * mm / mm
    )  # Use a dimensionless unit from pint if available, or just float


def test_steel_bar_initialization_imperial(steel_bar_grade60: SteelBar) -> None:
    """Test SteelBar initialization in imperial."""
    assert steel_bar_grade60.name == "Grade 60"
    assert_quantity_equal(steel_bar_grade60.f_y, 60 * ksi)
    assert_quantity_equal(
        steel_bar_grade60.E_s, 200 * GPa
    )  # E_s is hardcoded to 200 GPa
    # Convert E_s to psi for comparison: 200 GPa = 200e9 Pa = 29007550 psi ~ 29000 ksi
    # epsilon_y = 60 ksi / 29007.55 ksi = 0.002068
    expected_epsilon_y = (60 * ksi) / (200 * GPa).to("ksi")
    assert_quantity_equal(steel_bar_grade60.epsilon_y, expected_epsilon_y)


def test_steel_bar_get_properties(steel_bar_b500s: SteelBar) -> None:
    """Test get_properties for SteelBar."""
    props = steel_bar_b500s.get_properties()
    assert_quantity_equal(props["E_s"], Quantity(steel_bar_b500s.E_s.to("GPa")))
    assert_quantity_equal(props["f_y"], Quantity(steel_bar_b500s.f_y.to("MPa")))
    assert_quantity_equal(props["epsilon_y"], steel_bar_b500s.epsilon_y)


def test_steel_bar_str_representation(steel_bar_b500s: SteelBar) -> None:
    """Test __str__ method for SteelBar."""
    s = str(steel_bar_b500s)
    assert "SteelBar Properties (B500S):" in s
    assert f"  f_y: {steel_bar_b500s.f_y.to('MPa')}" in s
    assert f"  E_s: {steel_bar_b500s.E_s.to('GPa')}" in s
    assert f"  epsilon_y: {steel_bar_b500s.epsilon_y.magnitude:.4f}" in s
    assert f"  Density: {steel_bar_b500s.density}" in s


# --- Tests for SteelStrand ---


@pytest.fixture
def steel_strand_metric() -> SteelStrand:
    """Fixture for a typical SteelStrand (metric)."""
    return SteelStrand(name="Strand_1860", f_y=1600 * MPa)


def test_steel_strand_initialization_metric(steel_strand_metric: SteelStrand) -> None:
    """Test SteelStrand initialization in metric."""
    assert steel_strand_metric.name == "Strand_1860"
    assert_quantity_equal(steel_strand_metric.f_y, 1600 * MPa)
    assert_quantity_equal(steel_strand_metric.f_u, 1860 * MPa)  # Default
    assert_quantity_equal(steel_strand_metric.E_s, 190000 * MPa)  # Default
    assert_quantity_equal(steel_strand_metric.density, 7850 * kg / m**3)
    # epsilon_y = 1600 MPa / 190000 MPa = 0.008421...
    assert_quantity_equal(steel_strand_metric.epsilon_y, (1600 / 190000) * mm / mm)
    assert_quantity_equal(steel_strand_metric.prestress_stress, 0 * MPa)  # Default


def test_steel_strand_get_properties(steel_strand_metric: SteelStrand) -> None:
    """Test get_properties for SteelStrand."""
    props = steel_strand_metric.get_properties()
    assert_quantity_equal(props["E_s"], Quantity(steel_strand_metric.E_s.to("MPa")))
    assert_quantity_equal(props["f_y"], Quantity(steel_strand_metric.f_y.to("MPa")))
    assert_quantity_equal(props["f_u"], Quantity(steel_strand_metric.f_u.to("MPa")))


def test_steel_strand_str_representation(steel_strand_metric: SteelStrand) -> None:
    """Test __str__ method for SteelStrand."""
    s = str(steel_strand_metric)
    assert "SteelStrand Properties (Strand_1860):" in s
    assert f"  f_y: {steel_strand_metric.f_y.to('MPa')}" in s
    assert f"  f_u: {steel_strand_metric.f_u.to('MPa')}" in s
    assert f"  E_s: {steel_strand_metric.E_s.to('MPa')}" in s
    assert f"  Prestress Stress: {steel_strand_metric.prestress_stress}" in s
    assert f"  Density: {steel_strand_metric.density}" in s



from typing import List, Dict, Optional
from pandas import DataFrame
import pandas as pd
import copy
from collections import OrderedDict

from mento.material import (
    Concrete,
    SteelBar,
    Concrete_ACI_318_19,
    Concrete_EN_1992_2004,
)
from mento.forces import Forces
from mento.beam import RectangularBeam
from mento import mm, cm, kN, MPa, m, inch, ft, kNm
from mento.node import Node

pd.set_option("future.no_silent_downcasting", True)


class BeamSummary:
    def __init__(
        self, concrete: Concrete, steel_bar: SteelBar, beam_list: DataFrame
    ) -> None:
        self.concrete: Concrete = concrete
        self.steel_bar: SteelBar = steel_bar
        self.beam_list: DataFrame = beam_list
        self.units_row: List[str] = []
        self.data: DataFrame = None
        self.nodes: List[Node] = []
        self._beam_summary: List = []
        self.check_and_process_input()
        self.convert_to_nodes()

    def check_and_process_input(self) -> None:
        # Separate the header, units, and data
        self.units_row = self.beam_list.iloc[0].tolist()  # Second row (units)
        data = self.beam_list.iloc[
            1:
        ].copy()  # Data rows (after removing the units row)

        # Convert NaN in units to "dimensionless"
        self.units_row = ["" if pd.isna(unit) else unit for unit in self.units_row]

        # Validate the units row
        self.validate_units(self.units_row)

        # Convert NaN to 0 in the data rows
        data.iloc[:, 2:] = data.iloc[:, 2:].fillna(0).astype(float)
        # Convert specific columns to int and others to float
        columns_to_int = ["ns", "n1", "n2", "n3", "n4"]
        for col in columns_to_int:
            if col in data.columns:
                data[col] = data[col].astype(int)

        # Apply units to the corresponding columns, skipping the first column
        for i in range(
            1, len(self.units_row)
        ):  # Start from the second column (index 1)
            unit_str = self.units_row[i]
            if unit_str != "":
                unit = self.get_unit_variable(unit_str)
                if isinstance(data.iloc[:, i], pd.Series):
                    # data.iloc[:, i] = data.iloc[:, i].apply(lambda x: x * unit)
                    data.iloc[:, i] = data.iloc[:, i].apply(lambda x: x * unit)

        # Store the processed data
        self.data = data
        # print("Processed Data: Ok")

    def validate_units(self, units_row: List) -> None:
        valid_units = {"m", "mm", "cm", "inch", "ft", "kN", "kNm", ""}
        for unit_str in units_row:
            if unit_str and unit_str not in valid_units:
                raise ValueError(
                    f"Invalid unit '{unit_str}' detected. Allowed units: {valid_units}"
                )
        # print("Processed Units: Ok")

    def get_unit_variable(self, unit_str: str) -> Dict:
        # Map strings to actual unit variables (predefined in the script)
        unit_map = {
            "mm": mm,
            "cm": cm,
            "m": m,
            "in": inch,
            "ft": ft,
            "kN": kN,
            "kNm": kNm,
            "MPa": MPa,
        }
        if unit_str in unit_map:
            return unit_map[unit_str]
        else:
            raise ValueError(f"Unit '{unit_str}' is not recognized.")

    def convert_to_nodes(self) -> None:
        self.nodes = []

        for index, row in self.data.iterrows():
            # Extract forces for each row
            M_y = row["My"]
            N_x = row["Nx"]  # Positive for compression
            V_z = row["Vz"]
            comb = row["Comb."]

            # Ensure these are pint.Quantity objects with correct units
            forces = Forces(label=comb, M_y=M_y, N_x=N_x, V_z=V_z)

            # Extract geometric properties of the beam (width and height)
            width = row["b"]
            height = row["h"]
            c_c = row["cc"]

            # Create a rectangular concrete beam using the extracted values
            beam = RectangularBeam(
                label=row["Label"],
                concrete=self.concrete,
                steel_bar=self.steel_bar,
                width=width,
                height=height,
                c_c=c_c,
            )
            # Set transverse rebar (stirrups) for the beam
            n_stirrups = row["ns"]  # Number of stirrups
            d_b = row["dbs"]  # Diameter of rebar (mm)
            s_l = row["sl"]  # Spacing of stirrups (cm)

            if n_stirrups != 0:
                beam.set_transverse_rebar(n_stirrups=n_stirrups, d_b=d_b, s_l=s_l)

            # Set longitudinal rebar at the bottom if n1 is not 0
            n1 = row["n1"]
            d_b1 = row["db1"]  # Diameter in mm
            n2 = row["n2"]
            d_b2 = row["db2"]  # Diameter in mm
            n3 = row["n3"]
            d_b3 = row["db3"]  # Diameter in mm
            n4 = row["n4"]
            d_b4 = row["db4"]  # Diameter in mm
            if n1 != 0:
                if M_y >= 0 * kNm:
                    beam.set_longitudinal_rebar_bot(
                        n1, d_b1, n2, d_b2, n3, d_b3, n4, d_b4
                    )
                else:
                    beam.set_longitudinal_rebar_top(
                        n1, d_b1, n2, d_b2, n3, d_b3, n4, d_b4
                    )
            # Create a Node for each pair of beam and forces
            node = Node(section=beam, forces=forces)

            # Store the section and its corresponding forces
            self.nodes.append(node)

    def check(self, capacity_check: bool = False) -> DataFrame:
        """
        Perform a check on all beams in the summary.

        Parameters
        ----------
        capacity_check : bool, optional
            If True, resets all forces in the node to zero to perform a capacity check.
            Otherwise, uses the forces currently assigned to the node.

        Returns
        -------
        DataFrame
            A DataFrame with the results of the check.
        """

        results_list = []
        for node in self.nodes:
            beam: RectangularBeam = node.section  # type: ignore
            original_forces = [copy.deepcopy(force) for force in node.get_forces_list()]

            rebar_v = (
                "-"
                if beam._stirrup_n == 0
                else f"{int(beam._stirrup_n)}eØ{int(beam._stirrup_d_b.to('mm').magnitude)}/{int(beam._stirrup_s_l.to('cm').magnitude)}"
            )  # noqa: E501
            rebar_f_top = (
                "-"
                if beam._n1_t == 0
                else (
                    f"{beam._format_longitudinal_rebar_string(beam._n1_t, beam._d_b1_t, beam._n2_t, beam._d_b2_t)}"
                    + (
                        f" ++ {beam._format_longitudinal_rebar_string(beam._n3_t, beam._d_b3_t, beam._n4_t, beam._d_b4_t)}"
                        if beam._n3_t != 0
                        else ""
                    )
                )
            )
            rebar_f_bot = (
                "-"
                if beam._n1_b == 0
                else (
                    f"{beam._format_longitudinal_rebar_string(beam._n1_b, beam._d_b1_b, beam._n2_b, beam._d_b2_b)}"
                    + (
                        f" ++ {beam._format_longitudinal_rebar_string(beam._n3_b, beam._d_b3_b, beam._n4_b, beam._d_b4_b)}"
                        if beam._n3_b != 0
                        else ""
                    )
                )
            )

            if capacity_check:
                # Remove all forces assignments
                node.clear_forces()
                # Create empty force
                empty_force = Forces()
                node.add_forces(empty_force)
                # Perform the shear check
                shear_results = node.check_shear()
                flexure_results = node.check_flexure()
                # Common data
                common_data = {
                    "Beam": beam.label,
                    "b": beam.width.magnitude,
                    "h": beam.height.magnitude,
                    "As,top": rebar_f_top,
                    "As,bot": rebar_f_bot,
                    "Av": rebar_v,
                    "As,top,real": round(beam._A_s_top.to("cm**2").magnitude, 1),
                    "As,bot,real": round(beam._A_s_bot.to("cm**2").magnitude, 1),
                    "Av,real": round(shear_results["Av"][0], 1),
                }

                common_units = {
                    "Beam": "",
                    "b": "cm",
                    "h": "cm",
                    "As,top": "",
                    "As,bot": "",
                    "Av": "",
                    "As,top,real": "cm²",
                    "As,bot,real": "cm²",
                    "Av,real": "cm²/m",
                }

                # Code-specific data #TODO
                if isinstance(self.concrete, Concrete_EN_1992_2004):
                    code_specific_data = {
                        "MRd,top": round(beam._M_Rd_top.to("kN*m").magnitude, 1),
                        "MRd,bot": round(beam._M_Rd_bot.to("kN*m").magnitude, 1),
                        "VRd": shear_results["V_Rd"][0],
                    }
                    code_specific_units = {
                        "MRd,top": "kNm",
                        "MRd,bot": "kNm",
                        "VRd": "kN",
                    }
                else:  # ACI 318-19 or CIRSOC 201-25 design code
                    code_specific_data = {
                        "ØMn,top": round(beam._phi_M_n_top.to("kN*m").magnitude, 1),
                        "ØMn,bot": round(beam._phi_M_n_bot.to("kN*m").magnitude, 1),
                        "ØVn": shear_results["ØVn"][0],
                    }
                    code_specific_units = {
                        "ØMn,top": "kNm",
                        "ØMn,bot": "kNm",
                        "ØVn": "kN",
                    }

                # Merge data dictionaries
                merged_data = {**common_data, **code_specific_data}
                # Extract DCR values and remove them from the merged dictionary
                dcr_keys = ["DCRb,top", "DCRb,bot", "DCRv"]
                dcr_values = {key: merged_data.pop(key) for key in dcr_keys}
                # Rebuild results_dict with DCR values placed at the end
                results_dict = OrderedDict({**merged_data, **dcr_values})

                # Merge units dictionaries
                merged_units = {**common_units, **code_specific_units}

                # Extract and reorder DCR unit keys
                dcr_units = {key: merged_units.pop(key) for key in dcr_keys}

                # Build the units row with DCRs at the end
                units_row = pd.DataFrame([OrderedDict({**merged_units, **dcr_units})])
                # Restore the original forces after capacity check
                node.clear_forces()
                node.add_forces(original_forces)
            else:
                # Perform the shear check
                shear_results = (
                    node.check_shear().iloc[1:].reset_index(drop=True)
                )  # Skip the first row (units)
                flexure_results = (
                    node.check_flexure().iloc[1:].reset_index(drop=True)
                )  # Skip the first row (units)
                # Common data that doesn't change between codes
                common_data = {
                    "Beam": beam.label,
                    "b": int(beam.width.magnitude),
                    "h": int(beam.height.magnitude),
                    "cc": int(beam.c_c.magnitude),
                    "As,top": rebar_f_top,
                    "As,bot": rebar_f_bot,
                    "Av": rebar_v,
                    "As,req,top": round(beam._A_s_req_top.to("cm**2").magnitude, 2),
                    "As,req,bot": round(beam._A_s_req_bot.to("cm**2").magnitude, 2),
                    "Av,req": round(shear_results["Av,req"][0], 2),
                    "Av,real": round(shear_results["Av"][0], 2),
                    "DCRb,top": round(beam._DCRb_top, 2),
                    "DCRb,bot": round(beam._DCRb_bot, 2),
                    "DCRv": round(shear_results["DCR"][0], 2),
                }

                common_units = {
                    "Beam": "",
                    "b": "cm",
                    "h": "cm",
                    "cc": "mm",
                    "As,top": "",
                    "As,bot": "",
                    "Av": "",
                    "As,req,top": "cm²/m",
                    "As,req,bot": "cm²/m",
                    "Av,req": "cm²/m",
                    "Av,real": "cm²/m",
                    "DCRb,top": "",
                    "DCRb,bot": "",
                    "DCRv": "",
                }

                # Code-specific data
                if isinstance(self.concrete, Concrete_EN_1992_2004):  # TODO
                    code_specific_data = {
                        "MEd": round(flexure_results["Mu"][0], 1),
                        "VEd": round(shear_results["Vu"][0], 1),
                        "NEd": round(shear_results["Nu"][0], 1),
                        "MRd,top": round(beam._M_Rd_top.to("kN*m").magnitude, 1),
                        "MRd,bot": round(beam._M_Rd_bot.to("kN*m").magnitude, 1),
                        "VRd": round(shear_results["V_Rd"][0], 1),
                    }
                    code_specific_units = {
                        "MEd": "kNm",
                        "VEd": "kN",
                        "NEd": "kN",
                        "MRd,top": "kNm",
                        "MRd,bot": "kNm",
                        "VRd": "kN",
                    }
                else:  # ACI 318-19 or CIRSOC 201-25 design code
                    code_specific_data = {
                        "Mu": round(flexure_results["Mu"][0], 1),
                        "Vu": round(shear_results["Vu"][0], 1),
                        "Nu": round(shear_results["Nu"][0], 1),
                        "ØMn,top": round(beam._phi_M_n_top.to("kN*m").magnitude, 1),
                        "ØMn,bot": round(beam._phi_M_n_bot.to("kN*m").magnitude, 1),
                        "ØVn": round(shear_results["ØVn"][0], 1),
                    }
                    code_specific_units = {
                        "Mu": "kNm",
                        "Vu": "kN",
                        "Nu": "kN",
                        "ØMn,top": "kNm",
                        "ØMn,bot": "kNm",
                        "ØVn": "kN",
                    }

                # Merge data dictionaries
                merged_data = {**common_data, **code_specific_data}
                # Extract DCR values and remove them from the merged dictionary
                dcr_keys = ["DCRb,top", "DCRb,bot", "DCRv"]
                dcr_values = {key: merged_data.pop(key) for key in dcr_keys}
                # Rebuild results_dict with DCR values placed at the end
                results_dict = OrderedDict({**merged_data, **dcr_values})

                # Merge units dictionaries
                merged_units = {**common_units, **code_specific_units}

                # Extract and reorder DCR unit keys
                dcr_units = {key: merged_units.pop(key) for key in dcr_keys}

                # Build the units row with DCRs at the end
                units_row = pd.DataFrame([OrderedDict({**merged_units, **dcr_units})])

            # Add the results to the list
            results_list.append(results_dict)

        # Convert results list into a DataFrame
        results_df = pd.DataFrame(results_list)

        # Combine the units row with the results DataFrame
        final_df = pd.concat([units_row, results_df], ignore_index=True)
        return final_df

    def shear_results(
        self, index: Optional[int] = None, capacity_check: bool = False
    ) -> DataFrame:
        """
        Get shear results for one or all beams.
        Includes a units row only once at the top.
        """
        if index is not None:
            if index - 1 >= len(self.nodes):
                raise IndexError(f"Index {index} is out of range for the beam list.")
            node = self.nodes[max(index - 1, 0)]
            df = self._process_beam_for_check(node, "shear", capacity_check)
            units_row = node.section._get_units_row_shear()  # type: ignore
            df_all = pd.concat([units_row, df], ignore_index=True)
        else:
            # For all nodes, only include the units row once
            results = []
            units_row_added = False
            for item in self.nodes:
                df = self._process_beam_for_check(item, "shear", capacity_check)
                if not units_row_added:
                    units_row = item.section._get_units_row_shear()  # type: ignore
                    results.append(units_row)
                    units_row_added = True
                results.append(df)

            df_all = pd.concat(results, ignore_index=True)

        # Separate the units row (first row) from the data
        units_row = df_all.iloc[[0]]  # DataFrame with 1 row
        data_rows = df_all.iloc[1:].copy()

        # Try to round numeric columns (even if object dtype), coercing errors
        for col in data_rows.columns:
            data_rows[col] = pd.to_numeric(data_rows[col], errors="ignore")
            if pd.api.types.is_numeric_dtype(data_rows[col]):
                data_rows[col] = data_rows[col].round(2)

        # Recombine units + data
        df_final = pd.concat([units_row, data_rows], ignore_index=True)
        return df_final

    def _process_beam_for_check(
        self, node: Node, check_type: str, capacity_check: bool
    ) -> DataFrame:
        """
        Shared method to process beam for either shear or flexure checks.

        :param node: Node object to check
        :param check_type: Either 'shear' or 'flexure'
        :param capacity_check: If True, performs capacity check (resets forces)
        :return: Results DataFrame
        """
        original_forces = [copy.deepcopy(force) for force in node.get_forces_list()]

        if capacity_check:
            node.clear_forces()
            node.add_forces(Forces())  # Add empty force

        # Run the appropriate check
        if check_type == "shear":
            results = node.check_shear().iloc[1:].reset_index(drop=True)
        elif check_type == "flexure":
            results = node.check_flexure().iloc[1:].reset_index(drop=True)
        else:
            raise ValueError("check_type must be either 'shear' or 'flexure'")

        # Restore original forces if we did a capacity check
        if capacity_check:
            node.clear_forces()
            node.add_forces(original_forces)

        return results


def main() -> None:
    conc = Concrete_ACI_318_19(name="C25", f_c=25 * MPa)
    # conc = Concrete_EN_1992_2004(name="C25", f_c=25 * MPa)
    steel = SteelBar(name="ADN 500", f_y=500 * MPa)
    # input_df = pd.read_excel(r'.\tests\examples\Mento-Input.xlsx', sheet_name='Beams', usecols='B:R', skiprows=4)
    data = {
        "Label": ["", "V101", "V102", "V103", "V104"],
        "Comb.": ["", "ELU 1", "ELU 1", "ELU 1", "ELU 1"],
        "b": ["cm", 20, 20, 20, 20],
        "h": ["cm", 50, 50, 50, 50],
        "cc": ["mm", 25, 25, 25, 25],
        "Nx": ["kN", 0, 0, 0, 0],
        "Vz": ["kN", 20, -50, 100, 100],
        "My": ["kNm", 0, -35, 40, 45],
        "ns": ["", 0, 1.0, 1.0, 1.0],
        "dbs": ["mm", 0, 6, 6, 6],
        "sl": ["cm", 0, 20, 20, 20],
        "n1": ["", 2.0, 2, 2.0, 2.0],
        "db1": ["mm", 12, 12, 12, 12],
        "n2": ["", 1.0, 1, 1.0, 0.0],
        "db2": ["mm", 10, 16, 10, 0],
        "n3": ["", 2.0, 0.0, 2.0, 0.0],
        "db3": ["mm", 12, 0, 16, 0],
        "n4": ["", 0, 0.0, 0, 0.0],
        "db4": ["mm", 0, 0, 0, 0],
    }
    input_df = pd.DataFrame(data)
    # print(input_df)
    beam_summary = BeamSummary(concrete=conc, steel_bar=steel, beam_list=input_df)
    # print(beam_summary.data)
    # capacity = beam_summary.check(capacity_check=True)
    # print(capacity)
    check = beam_summary.check(capacity_check=False)
    print(check)
    # # beam_summary.check().to_excel('hola.xlsx', index=False)
    results = beam_summary.shear_results(capacity_check=False)
    # results = beam_summary.shear_results(capacity_check=True)
    print(results)
    # beam_summary.nodes[2].shear_results_detailed()


if __name__ == "__main__":
    main()



import pytest
import pandas as pd
from unittest.mock import (
    MagicMock,
    call,
)  # Import MagicMock and call for robust mocking
from typing import List, Generator

from mento.section import Section  # Assuming Section is correctly imported
from mento.forces import Forces  # Assuming Forces is correctly imported
from mento.node import Node
from mento.units import kN, m  # For creating sample Forces


# --- Fixtures ---


@pytest.fixture
def mock_section() -> MagicMock:
    """
    A mock Section object to pass to Node.
    Its methods will be mocked to track calls.
    We also mock `node` attribute, `label`, and `results` property.
    """
    mock_sec = MagicMock(
        spec=Section
    )  # spec=Section ensures mock has Section's attributes/methods
    mock_sec.node = None  # Node will set this
    mock_sec.label = "TestSection"  # For __repr__ test
    mock_sec.results = MagicMock()  # Mock the property itself
    mock_sec.check_flexure.return_value = pd.DataFrame({"Result": ["Flexure Check OK"]})
    mock_sec.design_flexure.return_value = pd.DataFrame(
        {"Result": ["Flexure Design Done"]}
    )
    mock_sec.check_shear.return_value = pd.DataFrame({"Result": ["Shear Check OK"]})
    mock_sec.design_shear.return_value = pd.DataFrame({"Result": ["Shear Design Done"]})
    mock_sec.shear_results_detailed.return_value = None  # Assuming no specific return
    mock_sec.shear_results_detailed_doc.return_value = None
    mock_sec.flexure_results_detailed.return_value = None
    mock_sec.flexure_results_detailed_doc.return_value = None
    return mock_sec


@pytest.fixture
def sample_force_single() -> Forces:
    """A single sample Forces object."""
    return Forces(N_x=100 * kN, V_z=10 * kN, M_y=20 * kN * m)


@pytest.fixture
def sample_forces_list() -> List[Forces]:
    """A list of sample Forces objects."""
    return [
        Forces(N_x=50 * kN, V_z=5 * kN, M_y=10 * kN * m),
        Forces(N_x=75 * kN, V_z=8 * kN, M_y=15 * kN * m),
    ]


@pytest.fixture(autouse=True)
def reset_node_id_counter() -> Generator[None, None, None]:
    """Fixture to reset the Node._last_id counter before each test."""
    original_last_id = Node._last_id
    Node._last_id = 0
    yield  # Allow the test to run
    Node._last_id = (
        original_last_id  # Restore original state if needed, though not critical here
    )


# --- Tests for Initialization ---


def test_node_initialization_with_single_force(
    mock_section: MagicMock, sample_force_single: Forces
) -> None:
    """Test Node initialization with a single Forces object."""
    node = Node(mock_section, sample_force_single)
    assert node.section is mock_section
    assert node.forces == [sample_force_single]
    assert node.section.node is node  # Check if section.node is set to this node


def test_node_initialization_with_list_of_forces(
    mock_section: MagicMock, sample_forces_list: List[Forces]
) -> None:
    """Test Node initialization with a list of Forces objects."""
    node = Node(mock_section, sample_forces_list)
    assert node.section is mock_section
    assert node.forces == sample_forces_list
    assert node.section.node is node


def test_node_initialization_invalid_section_type(sample_force_single: Forces) -> None:
    """Test Node initialization with an invalid section type."""
    with pytest.raises(TypeError, match="section must be an instance of Section"):
        Node("not a section", sample_force_single)  # type: ignore


def test_node_initialization_invalid_forces_type(mock_section: MagicMock) -> None:
    """Test Node initialization with an invalid forces type."""
    with pytest.raises(
        TypeError, match="forces must be an instance of Forces or a list of Forces"
    ):
        Node(mock_section, "not a force")  # type: ignore

    with pytest.raises(
        TypeError, match="forces must be an instance of Forces or a list of Forces"
    ):
        Node(mock_section, [Forces(N_x=1 * kN), "not a force"])  # type: ignore


# --- Tests for ID Management ---


def test_node_id_increment(
    mock_section: MagicMock, sample_force_single: Forces
) -> None:
    """Test that Node IDs increment correctly."""
    node1 = Node(mock_section, sample_force_single)
    node2 = Node(mock_section, sample_force_single)  # Create another node

    assert node1.id == 1
    assert node2.id == 2
    assert Node._last_id == 2  # Check the class counter


def test_node_id_read_only(
    mock_section: MagicMock, sample_force_single: Forces
) -> None:
    """Test that the 'id' property is read-only."""
    node = Node(mock_section, sample_force_single)
    with pytest.raises(AttributeError):
        node.id = 999  # type: ignore [misc]


# --- Tests for Force Management ---


def test_add_single_force(mock_section: MagicMock, sample_force_single: Forces) -> None:
    """Test adding a single force to the node."""
    node = Node(mock_section, [])  # Initialize with empty forces
    node.add_forces(sample_force_single)
    assert node.forces == [sample_force_single]


def test_add_list_of_forces(
    mock_section: MagicMock, sample_forces_list: List[Forces]
) -> None:
    """Test adding a list of forces to the node."""
    node = Node(mock_section, Forces(N_x=10 * kN))  # Start with one force
    node.add_forces(sample_forces_list)
    assert len(node.forces) == 1 + len(sample_forces_list)
    assert node.forces[1:] == sample_forces_list


def test_add_forces_invalid_type(mock_section: MagicMock) -> None:
    """Test adding forces with invalid type."""
    node = Node(mock_section, [])
    with pytest.raises(
        TypeError, match="forces must be an instance of Forces or a list of Forces"
    ):
        node.add_forces("invalid_force")  # type: ignore

    with pytest.raises(
        TypeError, match="forces must be an instance of Forces or a list of Forces"
    ):
        node.add_forces([Forces(N_x=1 * kN), "invalid"])  # type: ignore


def test_get_forces_list(
    mock_section: MagicMock, sample_forces_list: List[Forces]
) -> None:
    """Test getting the list of forces."""
    node = Node(mock_section, sample_forces_list)
    retrieved_forces = node.get_forces_list()
    assert retrieved_forces == sample_forces_list
    assert retrieved_forces is node.forces  # Should return the actual list, not a copy


def test_clear_forces(
    mock_section: MagicMock, sample_forces_list: List[Forces]
) -> None:
    """Test clearing all forces from the node."""
    node = Node(mock_section, sample_forces_list)
    assert len(node.forces) > 0
    node.clear_forces()
    assert node.forces == []


# --- Tests for Delegation to Section Methods ---


def test_check_flexure_delegation(
    mock_section: MagicMock, sample_forces_list: List[Forces]
) -> None:
    """Test that check_flexure calls section.check_flexure with correct forces."""
    node = Node(mock_section, sample_forces_list)
    result_df = node.check_flexure()
    mock_section.check_flexure.assert_called_once_with(sample_forces_list)
    assert isinstance(result_df, pd.DataFrame)
    assert result_df.equals(pd.DataFrame({"Result": ["Flexure Check OK"]}))


def test_design_flexure_delegation(
    mock_section: MagicMock, sample_forces_list: List[Forces]
) -> None:
    """Test that design_flexure calls section.design_flexure with correct forces."""
    node = Node(mock_section, sample_forces_list)
    result_df = node.design_flexure()
    mock_section.design_flexure.assert_called_once_with(sample_forces_list)
    assert isinstance(result_df, pd.DataFrame)
    assert result_df.equals(pd.DataFrame({"Result": ["Flexure Design Done"]}))


def test_check_shear_delegation(
    mock_section: MagicMock, sample_forces_list: List[Forces]
) -> None:
    """Test that check_shear calls section.check_shear with correct forces."""
    node = Node(mock_section, sample_forces_list)
    result_df = node.check_shear()
    mock_section.check_shear.assert_called_once_with(sample_forces_list)
    assert isinstance(result_df, pd.DataFrame)
    assert result_df.equals(pd.DataFrame({"Result": ["Shear Check OK"]}))


def test_design_shear_delegation(
    mock_section: MagicMock, sample_forces_list: List[Forces]
) -> None:
    """Test that design_shear calls section.design_shear with correct forces."""
    node = Node(mock_section, sample_forces_list)
    result_df = node.design_shear()
    mock_section.design_shear.assert_called_once_with(sample_forces_list)
    assert isinstance(result_df, pd.DataFrame)
    assert result_df.equals(pd.DataFrame({"Result": ["Shear Design Done"]}))


def test_shear_results_detailed_delegation(
    mock_section: MagicMock, sample_force_single: Forces
) -> None:
    """Test shear_results_detailed calls section.shear_results_detailed with correct force."""
    node = Node(mock_section, [sample_force_single])
    node.shear_results_detailed(sample_force_single)
    mock_section.shear_results_detailed.assert_called_once_with(sample_force_single)
    # Also test with None
    node.shear_results_detailed(None)
    mock_section.shear_results_detailed.assert_has_calls(
        [call(sample_force_single), call(None)]
    )


def test_shear_results_detailed_doc_delegation(
    mock_section: MagicMock, sample_forces_list: List[Forces]
) -> None:
    """Test shear_results_detailed_doc calls section.shear_results_detailed_doc."""
    node = Node(mock_section, sample_forces_list)
    node.shear_results_detailed_doc()
    mock_section.shear_results_detailed_doc.assert_called_once_with(
        None
    )  # Default to None


def test_flexure_results_detailed_delegation(
    mock_section: MagicMock, sample_force_single: Forces
) -> None:
    """Test flexure_results_detailed calls section.flexure_results_detailed."""
    node = Node(mock_section, [sample_force_single])
    node.flexure_results_detailed(sample_force_single)
    mock_section.flexure_results_detailed.assert_called_once_with(sample_force_single)


def test_flexure_results_detailed_doc_delegation(
    mock_section: MagicMock, sample_forces_list: List[Forces]
) -> None:
    """Test flexure_results_detailed_doc calls section.flexure_results_detailed_doc."""
    node = Node(mock_section, sample_forces_list)
    node.flexure_results_detailed_doc()
    mock_section.flexure_results_detailed_doc.assert_called_once_with(None)


# --- Tests for __repr__ ---


def test_node_repr_with_forces(
    mock_section: MagicMock, sample_force_single: Forces
) -> None:
    """Test the __repr__ method when forces are present."""
    node = Node(mock_section, sample_force_single)
    expected_repr_part1 = (
        f"Node ID: {node.id} - Section label: {mock_section.label}\nForces Applied:\n"
    )
    # Assuming sample_force_single.__str__ or __repr__ returns a sensible string
    expected_repr_part2 = f" - {str(sample_force_single)}"
    assert expected_repr_part1 in repr(node)
    assert expected_repr_part2 in repr(node)


def test_node_repr_without_forces(mock_section: MagicMock) -> None:
    """Test the __repr__ method when no forces are present."""
    node = Node(mock_section, [])
    expected_repr = f"Node ID: {node.id} - Section label: {mock_section.label}\nForces Applied:No forces applied"
    assert repr(node) == expected_repr



from dataclasses import dataclass, field
from typing import Dict, Optional
from pint import Quantity
from pint.facets.plain import PlainQuantity

from mento.units import kN, kNm


@dataclass
class Forces:
    """
    A class to represent the forces acting on a structural element.

    Attributes
    ----------
    N_x : float
        Axial force along the x-axis (default is 0 kN).
    V_z : float
        Shear force along the z-axis (default is 0 kN).
    M_y : float
        Bending moment about the y-axis (default is 0 kN*m).
    unit_system : str
        The unit system to use for displaying forces ('metric' or 'imperial').

    Methods
    -------
    get_forces() -> dict
        Returns the forces as a dictionary with keys 'N_x', 'V_z', and 'M_y'.
    set_forces() -> None
        Sets the forces of the object with the provided values.
    compare_to(other: 'Forces', by: str = 'V_z') -> bool
        Compares this force with another force based on a given attribute.
    """

    _id: int = field(init=False, repr=False)  # Instance ID, assigned internally
    _last_id: int = field(
        default=0, init=False, repr=False
    )  # Class variable to keep track of last assigned ID
    label: Optional[str] = None
    _N_x: Quantity = field(default=0 * kN)  # Ensure you use the correct unit type here
    _V_z: Quantity = field(default=0 * kN)
    _M_y: Quantity = field(default=0 * kNm)
    unit_system: str = field(default="metric")  # Add unit_system as a field

    def __init__(
        self,
        label: Optional[str] = None,
        N_x: Quantity = 0 * kN,
        V_z: Quantity = 0 * kN,
        M_y: Quantity = 0 * kNm,
        unit_system: str = "metric",
    ) -> None:
        # Increment the class variable for the next unique ID
        Forces._last_id += 1
        self._id = (
            Forces._last_id
        )  # Private ID assigned internally, unique per instance

        # Initialize the label
        self.label = label
        self.unit_system = unit_system  # Set the unit system

        # Set the forces upon initialization
        self.set_forces(N_x, V_z, M_y)

    @property
    def id(self) -> int:
        """Read-only property for accessing the unique ID of the instance."""
        return self._id

    @property
    def N_x(self) -> Quantity:
        """Axial force along the x-axis (default is 0 kN)."""
        if self.unit_system == "metric":
            return self._N_x.to("kN")
        else:
            return self._N_x.to("kip")

    @property
    def V_z(self) -> Quantity:
        """Shear force along the z-axis (default is 0 kN)."""
        if self.unit_system == "metric":
            return self._V_z.to("kN")
        else:
            return self._V_z.to("kip")

    @property
    def M_y(self) -> Quantity:
        """Bending moment about the y-axis (default is 0 kN*m)."""
        if self.unit_system == "metric":
            return self._M_y.to("kN*m")
        else:
            return self._M_y.to("ft*kip")

    def get_forces(self) -> Dict[str, Quantity]:
        """Returns the forces as a dictionary with keys 'N_x', 'V_z', and 'M_y'."""
        if self.unit_system == "metric":
            return {
                "N_x": self._N_x.to("kN"),
                "V_z": self._V_z.to("kN"),
                "M_y": self._M_y.to("kN*m"),
            }
        else:
            return {
                "N_x": self._N_x.to("kip"),
                "V_z": self._V_z.to("kip"),
                "M_y": self._M_y.to("ft*kip"),
            }

    def set_forces(
        self, N_x: Quantity = 0 * kN, V_z: Quantity = 0 * kN, M_y: Quantity = 0 * kNm
    ) -> None:
        """Sets the forces in the object."""
        self._N_x = N_x
        self._V_z = V_z
        self._M_y = M_y

    def compare_to(self, other: "Forces", by: str = "V_z") -> bool:
        """Compares this force with another force based on a selected attribute.

        Parameters
        ----------
        other : Forces
            Another Forces instance to compare with.
        by : str
            The attribute to compare by ('N_x', 'V_z', or 'M_y').

        Returns
        -------
        bool
            True if this force is greater than the other force by the selected attribute.
        """
        if by not in ["N_x", "V_z", "M_y"]:
            raise ValueError(
                "Comparison attribute must be one of 'N_x', 'V_z', or 'M_y'"
            )
        return getattr(self, by).magnitude > getattr(other, by).magnitude

    def __str__(self) -> str:
        return (
            f"Force ID: {self.id}, Label: {self.label}, "
            f"N_x: {self.N_x}, V_z: {self.V_z}, M_y: {self.M_y}"
        )



import pytest
from dataclasses import fields
from pint import Quantity

from mento.settings import BeamSettings, GLOBAL_BEAM_SETTINGS
from mento.units import mm, inch


# Helper to check if a quantity matches another, allowing for slight float differences
def assert_quantity_equal(q1: Quantity, q2: Quantity, rtol: float = 1e-9) -> None:
    """Asserts two Pint Quantities are numerically and unit-wise equal."""
    assert q1.units == q2.units, f"Units mismatch: {q1.units} vs {q2.units}"
    assert (
        pytest.approx(q1.magnitude, rel=rtol) == q2.magnitude
    ), f"Magnitudes mismatch: {q1.magnitude} vs {q2.magnitude}"


# --- Fixtures ---


@pytest.fixture
def metric_settings() -> BeamSettings:
    """Fixture for default metric BeamSettings."""
    return BeamSettings(unit_system="metric")


@pytest.fixture
def imperial_settings() -> BeamSettings:
    """Fixture for default imperial BeamSettings."""
    return BeamSettings(unit_system="imperial")


# --- Tests for Default Values ---


def test_metric_defaults_all_set(metric_settings: BeamSettings) -> None:
    """Test that all metric defaults are correctly applied when no custom values are given."""
    expected_defaults = (
        BeamSettings._metric_defaults
    )  # Access the class variable for comparison

    for field_info in fields(BeamSettings):
        if field_info.name in expected_defaults:  # Check only fields with defaults
            actual_value = getattr(metric_settings, field_info.name)
            expected_value = expected_defaults[field_info.name]

            if isinstance(expected_value, Quantity):
                assert_quantity_equal(actual_value, expected_value)
            else:  # For max_bars_per_layer (int)
                assert actual_value == expected_value

    assert metric_settings.unit_system == "metric"


def test_imperial_defaults_all_set(imperial_settings: BeamSettings) -> None:
    """Test that all imperial defaults are correctly applied when no custom values are given."""
    expected_defaults = BeamSettings._imperial_defaults

    for field_info in fields(BeamSettings):
        if field_info.name in expected_defaults:
            actual_value = getattr(imperial_settings, field_info.name)
            expected_value = expected_defaults[field_info.name]

            if isinstance(expected_value, Quantity):
                assert_quantity_equal(actual_value, expected_value)
            else:
                assert actual_value == expected_value

    assert imperial_settings.unit_system == "imperial"


def test_override_specific_metric_setting() -> None:
    """Test overriding a specific setting in metric system."""
    custom_clear_spacing = 30 * mm
    settings = BeamSettings(unit_system="metric", clear_spacing=custom_clear_spacing)
    assert_quantity_equal(settings.clear_spacing, custom_clear_spacing)
    # Ensure other settings remain default metric
    assert_quantity_equal(
        settings.stirrup_diameter_ini,
        BeamSettings._metric_defaults["stirrup_diameter_ini"],
    )


def test_override_specific_imperial_setting() -> None:
    """Test overriding a specific setting in imperial system."""
    custom_stirrup_diameter = 0.5 * inch
    settings = BeamSettings(
        unit_system="imperial", stirrup_diameter_ini=custom_stirrup_diameter
    )
    assert_quantity_equal(settings.stirrup_diameter_ini, custom_stirrup_diameter)
    # Ensure other settings remain default imperial
    assert_quantity_equal(
        settings.clear_spacing, BeamSettings._imperial_defaults["clear_spacing"]
    )


def test_custom_unit_system_with_no_defaults_set() -> None:
    """Test that if a unit system is specified without a corresponding default dict, it uses metric."""
    # Temporarily rename _metric_defaults to simulate a missing imperial_defaults scenario
    # (though in your code, imperial_defaults is always present).
    # This scenario would require a more complex mock/patch, or testing how it handles an unknown 'unit_system' string.
    # Given the current implementation, it will always fall back to metric if not "imperial".
    settings = BeamSettings(unit_system="unknown_system")
    assert settings.unit_system == "unknown_system"  # Unit system itself is set
    assert_quantity_equal(
        settings.clear_spacing, BeamSettings._metric_defaults["clear_spacing"]
    )
    # This implies that any system other than "imperial" defaults to metric. This is good to test.


# --- Tests for Validation ---


def test_max_bars_per_layer_validation_zero() -> None:
    """Test that max_bars_per_layer must be at least 1 (zero case)."""
    with pytest.raises(ValueError, match="max_bars_per_layer must be at least 1"):
        BeamSettings(unit_system="metric", max_bars_per_layer=0)


def test_max_bars_per_layer_validation_negative() -> None:
    """Test that max_bars_per_layer must be at least 1 (negative case)."""
    with pytest.raises(ValueError, match="max_bars_per_layer must be at least 1"):
        BeamSettings(unit_system="metric", max_bars_per_layer=-5)


def test_max_bars_per_layer_valid() -> None:
    """Test that a valid max_bars_per_layer does not raise an error."""
    try:
        settings = BeamSettings(unit_system="metric", max_bars_per_layer=1)
        assert settings.max_bars_per_layer == 1
        settings = BeamSettings(unit_system="metric", max_bars_per_layer=5)  # Default
        assert settings.max_bars_per_layer == 5
    except ValueError as e:
        pytest.fail(f"Valid max_bars_per_layer raised an error: {e}")


# --- Tests for __str__ method ---


def test_str_metric_defaults(metric_settings: BeamSettings) -> None:
    """Test the string representation for default metric settings."""
    s = str(metric_settings)
    expected_lines = [
        "clear_spacing: 25.00 mm",
        "stirrup_diameter_ini: 8.00 mm",
        "vibrator_size: 30.00 mm",
        "layers_spacing: 25.00 mm",
        "max_diameter_diff: 5.00 mm",
        "minimum_longitudinal_diameter: 8.00 mm",
        "max_bars_per_layer: 5",
    ]
    for line in expected_lines:
        assert line in s
    # Ensure no unit_system in output
    assert "unit_system" not in s


def test_str_imperial_defaults(imperial_settings: BeamSettings) -> None:
    """Test the string representation for default imperial settings."""
    s = str(imperial_settings)
    # Note: Pint's default for 3/8 inch might be 0.375 inch or similar.
    # Using 'in' for inches is common with '~' format.
    expected_lines = [
        "clear_spacing: 1.00 in",
        "stirrup_diameter_ini: 0.38 in",  # 3/8 is 0.375, .2f will round
        "vibrator_size: 1.25 in",
        "layers_spacing: 1.00 in",
        "max_diameter_diff: 0.25 in",
        "minimum_longitudinal_diameter: 0.38 in",
        "max_bars_per_layer: 5",
    ]
    for line in expected_lines:
        assert line in s
    assert "unit_system" not in s


def test_str_with_custom_values() -> None:
    """Test the string representation with some custom values."""
    settings = BeamSettings(
        unit_system="metric", clear_spacing=40 * mm, max_bars_per_layer=7
    )
    s = str(settings)
    assert "clear_spacing: 40.00 mm" in s
    assert "max_bars_per_layer: 7" in s
    # Check that other defaults are still present
    assert "stirrup_diameter_ini: 8.00 mm" in s
    assert "unit_system" not in s


# --- Tests for GLOBAL_BEAM_SETTINGS ---


def test_global_beam_settings_exists() -> None:
    """Test that GLOBAL_BEAM_SETTINGS is an instance of BeamSettings."""
    assert isinstance(GLOBAL_BEAM_SETTINGS, BeamSettings)


def test_global_beam_settings_is_metric_by_default(
    metric_settings: BeamSettings,
) -> None:
    """Test that GLOBAL_BEAM_SETTINGS uses metric defaults."""
    # Compare with a fresh default metric settings instance
    assert GLOBAL_BEAM_SETTINGS == metric_settings
    # Note: Dataclass equality works by comparing all fields.
    # If any internal _id or _last_id logic affects equality, this might need adjustment.
    # Given your BeamSettings is immutable and no _id logic, direct comparison is fine.
    assert GLOBAL_BEAM_SETTINGS.unit_system == "metric"
    assert_quantity_equal(GLOBAL_BEAM_SETTINGS.clear_spacing, 25 * mm)



import pytest
import matplotlib.pyplot as plt
from matplotlib.colors import to_rgba
from matplotlib.patches import Rectangle, FancyBboxPatch
import numpy as np
from typing import Generator

from mento.material import SteelBar, Concrete
from mento.units import inch, cm, mm, MPa
from mento.rectangular import RectangularSection  # Assuming this is the module name
from mento.settings import GLOBAL_BEAM_SETTINGS
from mento.results import CUSTOM_COLORS


@pytest.fixture()
def concrete_c25() -> Concrete:
    return Concrete("C25")


@pytest.fixture()
def steel_b500s() -> SteelBar:
    return SteelBar(name="B500S", f_y=500 * MPa)


@pytest.fixture()
def basic_rectangular_section(
    concrete_c25: Concrete, steel_b500s: SteelBar
) -> RectangularSection:
    """Fixture for a basic RectangularSection instance."""
    return RectangularSection(
        label="R101",
        concrete=concrete_c25,
        steel_bar=steel_b500s,
        c_c=25 * mm,
        width=30 * cm,  # Example dimensions
        height=50 * cm,
    )


@pytest.fixture()
def imperial_rectangular_section(
    concrete_c25: Concrete, steel_b500s: SteelBar
) -> Generator[RectangularSection, None, None]:
    """Fixture for a RectangularSection with imperial units set in concrete."""
    # Temporarily override unit system for concrete for this fixture
    # In a real app, you'd likely configure this via settings or concrete constructor
    concrete_c25.unit_system = "imperial"
    section = RectangularSection(
        label="R102",
        concrete=concrete_c25,
        steel_bar=steel_b500s,
        c_c=1 * inch,  # Example imperial cover
        width=12 * inch,  # Example imperial width
        height=24 * inch,
    )
    yield section  # Use yield to ensure cleanup happens after the test
    concrete_c25.unit_system = "metric"  # Reset for other tests if fixture is reused


# --- Tests for RectangularSection ---


def test_rectangular_section_initialization(
    basic_rectangular_section: RectangularSection,
) -> None:
    """Test RectangularSection specific initialization."""
    assert basic_rectangular_section.width == 30 * cm
    assert basic_rectangular_section.height == 50 * cm
    assert basic_rectangular_section.label == "R101"
    assert basic_rectangular_section.c_c == 25 * mm


def test_rectangular_section_geometric_properties(
    basic_rectangular_section: RectangularSection,
) -> None:
    """Test calculation of geometric properties (A_x, I_y, I_z)."""
    expected_A_x = (30 * cm) * (50 * cm)
    assert basic_rectangular_section.A_x == expected_A_x.to("cm**2")

    expected_I_y = (30 * cm) * (50 * cm) ** 3 / 12
    assert basic_rectangular_section.I_y == expected_I_y.to("cm**4")

    expected_I_z = (50 * cm) * (30 * cm) ** 3 / 12
    assert basic_rectangular_section.I_z == expected_I_z.to("cm**4")


def test_rectangular_section_settings_property(
    basic_rectangular_section: RectangularSection,
) -> None:
    """Test that the settings property returns GLOBAL_BEAM_SETTINGS."""
    assert basic_rectangular_section.settings is GLOBAL_BEAM_SETTINGS


# --- Test for the plot method ---


def test_rectangular_section_plot_smoke_test(
    basic_rectangular_section: RectangularSection,
) -> None:
    """
    Smoke test for the plot method.
    Simply calls the method and ensures no exceptions are raised.
    Closes the plot to prevent it from showing during tests.
    """
    try:
        basic_rectangular_section.plot()
        # Close the plot to prevent it from popping up
        plt.close()
    except Exception as e:
        pytest.fail(f"Plot method raised an exception: {e}")


def test_rectangular_section_plot_components(
    basic_rectangular_section: RectangularSection,
) -> None:
    """
    Test that the plot method adds the expected matplotlib components
    (main rectangle, stirrup rectangles, annotations).
    """
    basic_rectangular_section.plot()
    ax = basic_rectangular_section._ax  # Access the private _ax created by plot

    assert ax is not None, "Plot method should create and assign an Axes object."

    # Check for patches (rectangles, FancyBboxPatch)
    # Expecting: 1 main Rectangle, 2 FancyBboxPatches for stirrup
    # + potentially other internal patches depending on matplotlib's rendering
    # It's safer to check for specific types and properties if possible.

    # Filter for Rectangle and FancyBboxPatch
    rectangles = [p for p in ax.patches if isinstance(p, Rectangle)]
    fancy_bboxes = [p for p in ax.patches if isinstance(p, FancyBboxPatch)]

    assert (
        len(rectangles) >= 1
    ), "Expected at least one Rectangle patch (the main section)."
    assert (
        len(fancy_bboxes) == 2
    ), "Expected exactly two FancyBboxPatch objects for the stirrup."

    # Check for annotations (dimensions)
    # Annotations are stored in ax.texts for text, ax.lines or ax.artists for arrows
    # matplotlib.axes.Axes.annotate returns an Annotation object.
    # Text labels are found in ax.texts
    assert (
        len(ax.texts) == 4
    ), "Expected 4 text annotations for dimensions (width, height)."

    # You could add more specific checks, e.g.:
    # - Check the coordinates of the main rectangle:
    main_rect = next((p for p in rectangles if p.get_x() == 0 and p.get_y() == 0), None)
    assert main_rect is not None
    assert np.isclose(main_rect.get_width(), basic_rectangular_section.width.magnitude)
    assert np.isclose(
        main_rect.get_height(), basic_rectangular_section.height.magnitude
    )
    assert main_rect.get_edgecolor() == to_rgba(CUSTOM_COLORS["dark_gray"])

    # - Check stirrup patch properties (example, will need exact dimensions based on calculation)
    #   This gets more complex quickly, focusing on existence and basic properties is often enough.

    plt.close()  # Close the figure created by the plot method


def test_rectangular_section_plot_imperial_units_text(
    imperial_rectangular_section: RectangularSection,
) -> None:
    """
    Test that the plot method displays dimensions in imperial units
    when the concrete's unit system is set to 'imperial'.
    """
    imperial_rectangular_section.plot()
    ax = imperial_rectangular_section._ax
    assert ax is not None, "The plot method did not assign an Axes object to _ax."

    # Extract text labels
    text_labels = [t.get_text() for t in ax.texts]

    # Check if the labels contain "inch" (represented by "in" in pint's compact format)
    # Adjust based on how pint formats "inch" in your setup (e.g., "in", "inch", etc.)
    # "{:~P}" often uses common abbreviations.
    assert any(
        "in" in label for label in text_labels
    ), "Expected imperial unit 'in' in dimension text."
    assert (
        "{:.0f~P}".format(imperial_rectangular_section.width.to("inch")) in text_labels
    )
    assert (
        "{:.0f~P}".format(imperial_rectangular_section.height.to("inch")) in text_labels
    )

    plt.close()


def test_rectangular_section_plot_metric_units_text(
    basic_rectangular_section: RectangularSection,
) -> None:
    """
    Test that the plot method displays dimensions in metric units (cm)
    when the concrete's unit system is not 'imperial'.
    """
    basic_rectangular_section.plot()  # Ensure the plot method is called
    ax = basic_rectangular_section._ax
    assert ax is not None, "The plot method did not assign an Axes object to _ax."

    # Extract text labels
    text_labels = [t.get_text() for t in ax.texts]

    # Check if the labels contain "cm"
    assert any(
        "cm" in label for label in text_labels
    ), "Expected metric unit 'cm' in dimension text."
    assert "{:.0f~P}".format(basic_rectangular_section.width.to("cm")) in text_labels
    assert "{:.0f~P}".format(basic_rectangular_section.height.to("cm")) in text_labels

    plt.close()



from .EN_1992_2004_beam import (
    _check_shear_EN_1992_2004,  # noqa: F401
    _design_shear_EN_1992_2004,  # noqa: F401
)  # noqa: F401
from .ACI_318_19_beam import (
    _check_shear_ACI_318_19,  # noqa: F401
    _design_shear_ACI_318_19,  # noqa: F401
    _check_flexure_ACI_318_19,  # noqa: F401
    _design_flexure_ACI_318_19,  # noqa: F401
)  # noqa: F401
